r"""Solve a single-component, liquid-vapor, van der Waals system.

This example solves a single-component, liquid-vapor, van der Waals system as
described by Wheeler *et al.* :cite:`PhysRevE.82.051601`. The free energy for this
system takes the form,

.. math:: :label: eq:reactiveWetting:liquidVapor1D:freeEnergy

   f = - \frac{e \rho^2}{m^2} + \frac{R T}{m} \left( \ln \frac{\rho}{m - \bar{v} \rho} \right)

where :math:`\rho` is the density. This free energy supports a two phase
equilibrium with densities given by :math:`\rho^l` and :math:`\rho^v` in the
liquid and vapor phases, respectively. The densities are determined by solving
the following system of equations,

.. math::
   :label: eq:reactiveWetting:liquidVapor1D:pressureEquilibrium

   P \left( \rho^l \right) = P \left( \rho^v \right)

and

.. math::
   :label: eq:reactiveWetting:liquidVapor1D:chemicalPotentialEquilibrium

   \mu \left( \rho^l \right) = \mu \left( \rho^v \right)

where :math:`\mu` is the chemical potential,

.. math::
   :label: eq:reactiveWetting:liquidVapor1D:chemicalPotential

   \mu = \frac{\partial f}{\partial \rho}

and :math:`P` is the pressure,

.. math::
   :label: eq:reactiveWetting:liquidVapor1D:pressure

   P = \rho \mu - f

One choice of thermodynamic parameters that yields a relatively physical two phase system is

>>> molarWeight = 0.118
>>> ee = -0.455971
>>> gasConstant = 8.314
>>> temperature = 650.
>>> vbar = 1.3e-05

with equilibrium density values of

>>> liquidDensity = 7354.3402662299995
>>> vaporDensity = 82.855803327810008

The equilibrium densities are verified by substitution into Eqs.
:eq:`eq:reactiveWetting:liquidVapor1D:pressureEquilibrium` and
:eq:`eq:reactiveWetting:liquidVapor1D:chemicalPotentialEquilibrium`. Firstly,
Eqs. :eq:`eq:reactiveWetting:liquidVapor1D:freeEnergy`,
:eq:`eq:reactiveWetting:liquidVapor1D:chemicalPotential` and
:eq:`eq:reactiveWetting:liquidVapor1D:pressure` are defined as python functions,

>>> from fipy import CellVariable, Grid1D, TransientTerm, VanLeerConvectionTerm, DiffusionTerm, ImplicitSourceTerm, ConvectionTerm, CentralDifferenceConvectionTerm, Viewer
>>> from fipy.tools import numerix

>>> def f(rho):
...     return ee * rho**2 / molarWeight**2 + gasConstant * temperature * rho / molarWeight * \
...            numerix.log(rho / (molarWeight - vbar * rho))

>>> def mu(rho):
...     return 2 * ee * rho / molarWeight**2 + gasConstant * temperature / molarWeight * \
...            (numerix.log(rho / (molarWeight - vbar * rho)) + molarWeight / (molarWeight - vbar * rho))

>>> def P(rho):
...     return rho * mu(rho) - f(rho)

The equilibrium densities values are verified with

>>> print(numerix.allclose(mu(liquidDensity), mu(vaporDensity)))
True

and

>>> print(numerix.allclose(P(liquidDensity), P(vaporDensity)))
True

In order to derive governing equations, the free energy functional is defined.

.. math::

   F = \int \left[ f + \frac{\epsilon T}{2} \left( \partial_j \rho \right)^2 \right] dV

Using standard dissipation laws, we write the governing equations for mass and
momentum conservation,

.. math::
   :label: eq:reactiveWetting:liquidVapor1D:mass

   \frac{\partial \rho}{\partial t} + \partial_j \left(\rho u_j \right) = 0

and

.. math::
   :label: eq:reactiveWetting:liquidVapor1D:momentum

   \frac{\partial \left( \rho u_i\right) }{\partial t} + \partial_j\left( \rho
   u_iu_j \right) = \partial_j\left( \nu \left[ \partial_j u_i + \partial_i u_j
   \right] \right) - \rho \partial_i \mu^{NC}

where the non-classical potential, :math:`\mu^{NC}`, is given by,

.. math::
  :label: eq:reactiveWetting:liquidVapor1D:nonClassicalPotential

   \mu^{NC} = \frac{\delta F}{\delta \rho} = \mu - \epsilon T \partial_j^2 \rho

As usual, to proceed, we define a mesh

>>> Lx = 1e-6
>>> nx = 100
>>> dx = Lx / nx
>>> mesh = Grid1D(nx=nx, dx=dx)

and the independent variables.

>>> density = CellVariable(mesh=mesh, hasOld=True, name=r'$\rho$')
>>> velocity = CellVariable(mesh=mesh, hasOld=True, name=r'$u$')
>>> densityPrevious = density.copy()
>>> velocityPrevious = velocity.copy()

The system of equations is solved in a fully coupled manner using a block
matrix. Defining :math:`\mu^{NC}` as an independent variable makes it easier to
script the equations without using higher order terms.

>>> potentialNC = CellVariable(mesh=mesh, name=r'$\mu^{NC}$')

>>> epsilon = 1e-16
>>> freeEnergy = (f(density) + epsilon * temperature / 2 * density.grad.mag**2).cellVolumeAverage

In order to solve the equations numerically, an interpolation method is used to
prevent the velocity and density fields decoupling. The following velocity
correction equation (expressed in discretized form) prevents decoupling from
occurring,

.. math::
   :label: eq:reactiveWetting:liquidVapor1D:correction

   u_{i,f}^c = \frac{A_f d_f}{\overline{d}_f} \left( \overline{ \rho \partial_i
   \mu^{NC} }_f - \overline{\rho}_f \partial_{i,f} \mu^{NC} \right)

where :math:`A_f` is the face area, :math:`d_f` is the distance between the
adjacent cell centers and :math:`\overline{a}_f` is the momentum conservation
equation's matrix diagonal. The overbar refers to an averaged value between the
two adjacent cells to the face. The notation :math:`\partial_{i,f}` refers to a
derivative evaluated directly at the face (not averaged). The variable
:math:`u_i^c` is used to modify the velocity used in
Eq. :eq:`eq:reactiveWetting:liquidVapor1D:mass` such that,

.. math::
   :label: eq:reactiveWetting:liquidVapor1D:massCorrected

   \frac{\partial \rho}{\partial t} + \partial_j \left(\rho \left[u_j + u_i^c
   \right] \right) = 0

Equation :eq:`eq:reactiveWetting:liquidVapor1D:massCorrected` becomes

>>> matrixDiagonal = CellVariable(mesh=mesh, name=r'$a_f$', value=1e+20, hasOld=True)
>>> correctionCoeff = mesh._faceAreas * mesh._cellDistances / matrixDiagonal.faceValue
>>> massEqn = TransientTerm(var=density) \
...           + VanLeerConvectionTerm(coeff=velocity.faceValue + correctionCoeff \
...                                         * (density * potentialNC.grad).faceValue, \
...                                   var=density) \
...           - DiffusionTerm(coeff=correctionCoeff * density.faceValue**2, var=potentialNC)

where the first term on the LHS of
Eq. :eq:`eq:reactiveWetting:liquidVapor1D:correction` is calculated in an
explicit manner in the ``VanLeerConvectionTerm`` and the second term is
calculated implicitly as a ``DiffusionTerm`` with :math:`\mu^{NC}` as the
independent variable.

In order to write Eq. :eq:`eq:reactiveWetting:liquidVapor1D:momentum` as a
:term:`FiPy` expression, the last term is rewritten such that,

.. math::

   \rho \partial_i \mu^{NC} = \partial_i \left( \rho \mu^{NC} \right) - \mu^{NC} \partial_i \rho

which results in

>>> viscosity = 1e-3
>>> ConvectionTerm = CentralDifferenceConvectionTerm
>>> momentumEqn = TransientTerm(coeff=density, var=velocity) \
...               + ConvectionTerm(coeff=[[1]] * density.faceValue * velocity.faceValue, var=velocity) \
...               == DiffusionTerm(coeff=2 * viscosity, var=velocity) \
...               - ConvectionTerm(coeff=density.faceValue * [[1]], var=potentialNC) \
...               + ImplicitSourceTerm(coeff=density.grad[0], var=potentialNC)

The only required boundary condition eliminates flow in or out of the domain.

>>> velocity.constrain(0, mesh.exteriorFaces)

As previously stated, the :math:`\mu^{NC}` variable will be solved implicitly. To
do this the Eq. :eq:`eq:reactiveWetting:liquidVapor1D:nonClassicalPotential` is
linearized in :math:`\rho` such that

.. math::
   :label: eq:reactiveWetting:liquidVapor1D:chemicalPotentialEquation

   \mu^{NC} = \mu^* + \left( \frac{\partial \mu}{\partial \rho} \right)^* \left( \rho - \rho^* \right) - \epsilon T \partial_j^2 \rho

The :math:`^*` superscript denotes the current held value. In :term:`FiPy`,
:math:`\frac{\partial \mu}{\partial \rho}` is written as,

>>> potentialDerivative = 2 * ee / molarWeight**2 + gasConstant * temperature * molarWeight / density / (molarWeight - vbar * density)**2

and :math:`\mu^*` is simply,

>>> potential = mu(density)

Eq. :eq:`eq:reactiveWetting:liquidVapor1D:chemicalPotentialEquation` can be
scripted as

>>> potentialNCEqn = ImplicitSourceTerm(coeff=1, var=potentialNC) \
...                  == potential \
...                  + ImplicitSourceTerm(coeff=potentialDerivative, var=density) \
...                  - potentialDerivative * density \
...                  - DiffusionTerm(coeff=epsilon * temperature, var=density)

Due to a quirk in :term:`FiPy`, the gradient of :math:`\mu^{NC}` needs to be
constrained on the boundary.  This is because ``ConvectionTerm``'s will
automatically assume a zero flux, which is not what we need in this case.

>>> potentialNC.faceGrad.constrain(value=[0], where=mesh.exteriorFaces)

All three equations are defined and an are combined together with

>>> coupledEqn = massEqn & momentumEqn & potentialNCEqn

The system will be solved as a phase separation problem with an initial density
close to the average density, but with some small amplitude noise. Under these
circumstances, the final condition should be two separate phases of roughly equal
volume. The initial condition for the density is defined by

>>> numerix.random.seed(2011)
>>> density[:] = (liquidDensity + vaporDensity) / 2 * \
...    (1  + 0.01 * (2 * numerix.random.random(mesh.numberOfCells) - 1))

Viewers are also defined.

>>> from fipy import input
>>> if __name__ == '__main__':
...     viewers = Viewer(density), Viewer(velocity), Viewer(potentialNC)
...     for viewer in viewers:
...         viewer.plot()
...     input('Arrange viewers, then press <return> to proceed...')
...     for viewer in viewers:
...         viewer.plot()

The following section defines the required control parameters. The `cfl`
parameter limits the size of the time step so that 
`dt = cfl * dx / max(velocity)`.

>>> cfl = 0.1
>>> tolerance = 1e-1
>>> dt = 1e-14
>>> timestep = 0
>>> relaxation = 0.5
>>> if __name__ == '__main__':
...     totalSteps = 1e10
... else:
...     totalSteps = 10

In the following time stepping scheme a time step is recalculated if the residual
increases between sweeps or the required tolerance is not attained within 20
sweeps. The major quirk in this scheme is the requirement of updating the
``matrixDiagonal`` using the entire coupled matrix. This could be achieved more
elegantly by calling ``cacheMatrix()`` only on the necessary part of the
equation. This currently doesn't work properly in :term:`FiPy`.

Beginning with :term:`FiPy` 4, solver :ref:`CONVERGENCE`
tolerance is normalized by the magnitude of the right-hand-side (RHS)
vector.  For this particular problem, the initial residual is much smaller
than the RHS and so the solver gets "stuck".  Changing the normalization to
use the initial residual at the beginning of each sweep allows the solution
to progress.  Another option would be to scale the tolerance appropriately,
but the value is so small (:math:`\sim 10^{-22}`) that this results in
underflow issues for some solvers.

>>> solver = coupledEqn.getDefaultSolver(criterion="initial", tolerance=1e-12)

.. note::

   :ref:`PETSC` intrinsically wants to use `"preconditioned"`
   normalization, which prevents the solver from getting "stuck" because
   the preconditioner effectively scales the RHS to be similar in magnitude
   to the residual.  Unfortunately, this normalization is not available for
   the other solver suites, so we don't use it as the default.

>>> while timestep < totalSteps:
... 
...     sweep = 0
...     dt *= 1.1
...     residual = 1.
...     initialResidual = None
... 
...     density.updateOld()
...     velocity.updateOld()
...     matrixDiagonal.updateOld()
... 
...     while residual > tolerance:
... 
...         densityPrevious[:] = density
...         velocityPrevious[:] = velocity
...         previousResidual = residual
... 
...         dt = min(dt, dx / max(abs(velocity)) * cfl)
... 
...         coupledEqn.cacheMatrix()
...         residual = coupledEqn.sweep(dt=dt, solver=solver)
...
...         if initialResidual is None:
...             initialResidual = residual
... 
...         residual = residual / initialResidual
...
...         if residual > previousResidual * 1.1 or sweep > 20:
...             density[:] = density.old
...             velocity[:] = velocity.old
...             matrixDiagonal[:] = matrixDiagonal.old
...             dt = dt / 10.
...             if __name__ == '__main__':
...                 print('Recalculate the time step')
...             timestep -= 1
...             break
...         else:
...             matrixDiagonal[:] = coupledEqn.matrix.takeDiagonal()[mesh.numberOfCells:2 * mesh.numberOfCells]
...             density[:] = relaxation * density + (1 - relaxation) * densityPrevious
...             velocity[:] = relaxation * velocity + (1 - relaxation) * velocityPrevious
... 
...         sweep += 1
... 
...     if __name__ == '__main__' and timestep % 10 == 0:
...         print('timestep: %e / %e, dt: %1.5e, free energy: %1.5e' % (timestep, totalSteps, dt, freeEnergy))
...         for viewer in viewers:
...             viewer.plot()
... 
...     timestep += 1

>>> from fipy import input
>>> if __name__ == '__main__':
...     input('finished')

>>> print(freeEnergy < 1.5e9)
True

.. .. bibmissing:: /refs.bib
    :sort:
"""

__docformat__ = 'restructuredtext'

if __name__ == '__main__':
    import fipy.tests.doctestPlus
    exec(fipy.tests.doctestPlus._getScript())



from petsc4py import PETSc

from ..convergence import (Convergence,
                           AbsoluteToleranceConvergence,
                           HappyBreakdownConvergence,
                           IteratingConvergence,
                           IterationConvergence,
                           RelativeToleranceConvergence,
                           Divergence,
                           BreakdownDivergence,
                           IllConditionedDivergence,
                           IterationDivergence,
                           NullDivergence,
                           OutOfRangeDivergence,
                           PreconditioningDivergence,
                           IllConditionedPreconditionerDivergence,
                           ToleranceDivergence)

# "The values KSP_CONVERGED_CG_NEG_CURVE, KSP_CONVERGED_CG_CONSTRAINED, and
# KSP_CONVERGED_STEP_LENGTH are returned only by the special KSPNASH,
# KSPSTCG, and KSPGLTR solvers which are used by the SNESNEWTONTR (trust
# region) solver."

class KSP_AbsoluteToleranceConvergence(AbsoluteToleranceConvergence):
    """Residual 2-norm less than abstol"""
    status_code = PETSc.KSP.ConvergedReason.CONVERGED_ATOL
    status_name = "KSP_CONVERGED_ATOL"
    suite = "petsc"

class KSP_RelativeToleranceConvergence(RelativeToleranceConvergence):
    """Residual 2-norm decreased by a factor of `rtol`, from 2-norm of right
    hand side.
    """
    status_code = PETSc.KSP.ConvergedReason.CONVERGED_RTOL
    status_name = "KSP_CONVERGED_RTOL"
    suite = "petsc"

class KSP_NormalAbsoluteToleranceConvergence(KSP_AbsoluteToleranceConvergence):
    if hasattr(PETSc.KSP.ConvergedReason, "CONVERGED_ATOL_NORMAL_EQUATIONS"):
        status_code = PETSc.KSP.ConvergedReason.CONVERGED_ATOL_NORMAL_EQUATIONS
    else:
        # CONVERGED_ATOL_NORMAL was deprecated in PETSc 3.24
        status_code = PETSc.KSP.ConvergedReason.CONVERGED_ATOL_NORMAL
    status_name = "KSP_CONVERGED_ATOL_NORMAL"

class KSP_NormalRelativeToleranceConvergence(KSP_RelativeToleranceConvergence):
    if hasattr(PETSc.KSP.ConvergedReason, "CONVERGED_RTOL_NORMAL_EQUATIONS"):
        status_code = PETSc.KSP.ConvergedReason.CONVERGED_RTOL_NORMAL_EQUATIONS
    else:
        # CONVERGED_RTOL_NORMAL was deprecated in PETSc 3.24
        status_code = PETSc.KSP.ConvergedReason.CONVERGED_RTOL_NORMAL
    status_name = "KSP_CONVERGED_RTOL_NORMAL"

class KSP_IterationConvergence(IterationConvergence):
    """Used by the `KSPPREONLY` solver after the single iteration of the
    preconditioner is applied.  Also used when the `KSPConvergedSkip()`
    convergence test routine is set in `KSP`.
    """
    status_code = PETSc.KSP.ConvergedReason.CONVERGED_ITS
    status_name = "KSP_CONVERGED_ITS"
    suite = "petsc"

class KSP_HappyBreakdownConvergence(HappyBreakdownConvergence):
    status_code = PETSc.KSP.ConvergedReason.CONVERGED_HAPPY_BREAKDOWN
    status_name = "KSP_CONVERGED_HAPPY_BREAKDOWN"
    suite = "petsc"

class KSP_IteratingConvergence(IteratingConvergence):
    """This flag is returned if you call `KSPGetConvergedReason()` while the
    `KSPSolve()` is still running.
    """
    status_code = PETSc.KSP.ConvergedReason.CONVERGED_ITERATING
    status_name = "KSP_CONVERGED_ITERATING"
    suite = "petsc"


class KSP_IterationDivergence(IterationDivergence):
    """Ran out of iterations before any convergence criteria was reached"""
    status_code = PETSc.KSP.ConvergedReason.DIVERGED_MAX_IT
    status_name = "KSP_DIVERGED_ITS"
    suite = "petsc"

class KSP_NullDivergence(NullDivergence):
    status_code = PETSc.KSP.ConvergedReason.DIVERGED_NULL
    status_name = "KSP_DIVERGED_NULL"
    suite = "petsc"

class KSP_ToleranceDivergence(ToleranceDivergence):
    """Residual norm increased by a factor of `divtol`.
    """
    status_code = PETSc.KSP.ConvergedReason.DIVERGED_DTOL
    status_name = "KSP_DIVERGED_DTOL"
    suite = "petsc"

class KSP_BreakdownDivergence(BreakdownDivergence):
    """Generic breakdown in method."""
    status_code = PETSc.KSP.ConvergedReason.DIVERGED_BREAKDOWN
    status_name = "KSP_DIVERGED_BREAKDOWN"
    suite = "petsc"

class KSP_BreakdownBICGDivergence(KSP_BreakdownDivergence):
    """Initial residual is orthogonal to preconditioned initial residual.
    Try a different preconditioner, or a different initial Level.)
    """
    status_code = PETSc.KSP.ConvergedReason.DIVERGED_BREAKDOWN_BICG
    status_name = "KSP_DIVERGED_BREAKDOWN_BICG"

class KSP_IndefiniteMatrixDivergence(IllConditionedDivergence):
    status_code = PETSc.KSP.ConvergedReason.DIVERGED_INDEFINITE_MAT
    status_name = "KSP_DIVERGED_INDEFINITE_MAT"
    suite = "petsc"

class KSP_NonSymmetricDivergence(IllConditionedDivergence):
    """It appears the operator or preconditioner is not symmetric and this
    Krylov method (`KSPCG`, `KSPMINRES`, `KSPCR`) requires symmetry
    """
    status_code = PETSc.KSP.ConvergedReason.DIVERGED_NONSYMMETRIC
    status_name = "KSP_DIVERGED_NONSYMMETRIC"
    suite = "petsc"

class KSP_PreconditioningDivergence(PreconditioningDivergence):
    """It was not possible to build or use the requested preconditioner.
    This is usually due to a zero pivot in a factorization.  It can also
    result from a failure in a subpreconditioner inside a nested
    preconditioner such as `PCFIELDSPLIT`.
    """
    status_code = PETSc.KSP.ConvergedReason.DIVERGED_PCSETUP_FAILED
    status_name = "KSP_DIVERGED_PC_FAILED"
    suite = "petsc"

class KSP_IndefinitePreconditionerDivergence(IllConditionedPreconditionerDivergence):
    """It appears the preconditioner is indefinite (has both positive and
    negative eigenvalues) and this Krylov method (`KSPCG`) requires it to be
    positive definite.
    """
    status_code = PETSc.KSP.ConvergedReason.DIVERGED_INDEFINITE_PC
    status_name = "KSP_DIVERGED_INDEFINITE_PC"
    suite = "petsc"

class KSP_NanOrInfDivergence(OutOfRangeDivergence):
    """Residual norm became Not-a-number or Inf likely due to 0/0."""
    status_code = PETSc.KSP.ConvergedReason.DIVERGED_NANORINF
    status_name = "KSP_DIVERGED_NANORINF"
    suite = "petsc"



r"""Create a level set variable in one dimension.

The level set
variable calculates its value over the domain to be the distance from
the zero level set. This can be represented succinctly in the
following equation with a boundary condition at the zero level set
such that,

.. math::

   \frac{\partial \phi}{\partial x} = 1

with the boundary condition, :math:`\phi = 0` at :math:`x = L / 2`.

The solution to this problem will be demonstrated in the following
script. Firstly, setup the parameters.

>>> from fipy import CellVariable, Grid1D, DistanceVariable, TransientTerm, FirstOrderAdvectionTerm, AdvectionTerm, Viewer
>>> from fipy.tools import numerix, serialComm

>>> dx = 0.5
>>> nx = 10

Construct the mesh.

.. index::
   single: Grid2D

>>> mesh = Grid1D(dx=dx, nx=nx, communicator=serialComm)

Construct a `distanceVariable` object.

>>> var = DistanceVariable(name='level set variable',
...                        mesh=mesh,
...                        value=-1.,
...                        hasOld=1)
>>> x = mesh.cellCenters[0]
>>> var.setValue(1, where=x > dx * nx / 2)

Once the initial positive and negative regions have been initialized
the `calcDistanceFunction()` method can be used to recalculate `var`
as a distance function from the zero level set.

>>> var.calcDistanceFunction() #doctest: +LSM

The problem can then be solved by executing the :meth:`~fipy.terms.term.Term.solve`
method of the equation.

>>> if __name__ == '__main__':
...     viewer = Viewer(vars=var, datamin=-5., datamax=5.)
...     viewer.plot()

The result can be tested with the following commands.

>>> print(numerix.allclose(var, x - dx * nx / 2)) #doctest: +LSM
1

"""
__docformat__ = 'restructuredtext'

from fipy import input

if __name__ == '__main__':
    import fipy.tests.doctestPlus
    exec(fipy.tests.doctestPlus._getScript())
    input("finished")




r"""Solve a phase-field (Allen-Cahn) problem in one-dimension.

To run this example from the base FiPy directory, type
``python examples/phase/simple/input.py`` at the command line.  A viewer
object should appear and, after being prompted to step through the different
examples, the word ``finished`` in the terminal.

This example takes the user through assembling a simple problem with
FiPy.  It describes a steady 1D phase field problem with no-flux
boundary conditions such that,

.. math::
   :label: eq-phase:simple

   \frac{1}{M_\phi}\frac{\partial \phi}{\partial t} =
   \kappa_\phi \nabla^2\phi
   - \frac{\partial f}{\partial \phi}

For solidification problems, the Helmholtz free energy is frequently
given by

.. math::

   f(\phi,T) = \frac{W}{2}g(\phi) + L_v\frac{T-T_M}{T_M}p(\phi)

where :math:`W` is the double-well barrier height between phases, :math:`L_v` is the latent
heat, :math:`T` is the temperature, and :math:`T_M` is the melting point.

One possible choice for the double-well function is

.. math::

   g(\phi) = \phi^2(1 - \phi)^2

and for the interpolation function is

.. math::

   p(\phi) = \phi^3(6\phi^2 - 15\phi + 10).

We create a 1D solution mesh

>>> from fipy import CellVariable, Variable, Grid1D, DiffusionTerm, TransientTerm, ImplicitSourceTerm, DummySolver, Viewer
>>> from fipy.tools import numerix

>>> L = 1.
>>> nx = 400
>>> dx = L / nx

.. index::
   single: Grid1D

>>> mesh = Grid1D(dx = dx, nx = nx)

We create the phase field variable

.. index::
   single: CellVariable

>>> phase = CellVariable(name = "phase",
...                      mesh = mesh)

and set a step-function initial condition

.. math::

   \phi =
   \begin{cases}
       1 & \text{for $x \le L/2$} \\
       0 & \text{for $x > L/2$}
   \end{cases}
   \quad\text{at $t = 0$}

>>> x = mesh.cellCenters[0]
>>> phase.setValue(1.)
>>> phase.setValue(0., where=x > L/2)

If we are running interactively, we'll want a viewer to see the results

.. index::
   pair: module; fipy.viewers

>>> from fipy import input
>>> if __name__ == '__main__':
...     viewer = Viewer(vars = (phase,))
...     viewer.plot()
...     input("Initial condition. Press <return> to proceed...")

.. image:: /figures/examples/phase/simple/step.*
   :width: 50%
   :align: center
   :alt: step-function initial condition

We choose the parameter values,

>>> kappa = 0.0025
>>> W = 1.
>>> Lv = 1.
>>> Tm = 1.
>>> T = Tm
>>> enthalpy = Lv * (T - Tm) / Tm

We build the equation by assembling the appropriate terms.  Since, with
:math:`T = T_M`
we are interested in a steady-state solution, we omit the transient term
:math:`(1/M_\phi)\frac{\partial \phi}{\partial t}`.

The analytical solution for this steady-state phase field problem, in an infinite domain, is

.. math::
   :label: eq-phase:simple:analytical

   \phi = \frac{1}{2}\left[1 - \tanh\frac{x-L/2}{2\sqrt{\kappa/W}}\right]

or

.. index::
   single: tanh
   single: sqrt

>>> x = mesh.cellCenters[0]
>>> analyticalArray = 0.5*(1 - numerix.tanh((x - L/2)/(2*numerix.sqrt(kappa/W))))

We treat the diffusion term
:math:`\kappa_\phi \nabla^2\phi`
implicitly,

.. note::

   "Diffusion" in :term:`FiPy` is not limited to the movement of atoms, but
   rather refers to the spontaneous spreading of any quantity (e.g.,
   solute, temperature, or in this case "phase") by flow "down" the
   gradient of that quantity.

The source term is

.. math::

   S =
   -\frac{\partial f}{\partial \phi} &= -\frac{W}{2}g'(\phi) - L\frac{T-T_M}{T_M}p'(\phi) \\
   &= -\left[W\phi(1-\phi)(1-2\phi) + L\frac{T-T_M}{T_M}30\phi^2(1-\phi)^2\right] \\
   &= m_\phi\phi(1-\phi)

where
:math:`m_\phi \equiv -[W(1-2\phi) + 30\phi(1-\phi)L\frac{T-T_M}{T_M}]`.

The simplest approach is to add this source explicitly

>>> mPhi = -((1 - 2 * phase) * W + 30 * phase * (1 - phase) * enthalpy)
>>> S0 = mPhi * phase * (1 - phase)
>>> eq = S0 + DiffusionTerm(coeff=kappa)

After solving this equation

>>> eq.solve(var = phase, solver=DummySolver())

we obtain the surprising result that :math:`\phi` is zero everywhere.

>>> print(phase.allclose(analyticalArray, rtol = 1e-4, atol = 1e-4))
0
>>> from fipy import input
>>> if __name__ == '__main__':
...     viewer.plot()
...     input("Fully explicit source. Press <return> to proceed...")

.. image:: /figures/examples/phase/simple/explicit.*
   :width: 50%
   :align: center
   :alt: steady-state phase field zero everywhere

On inspection, we can see that this occurs because, for our step-function initial condition,
:math:`m_\phi = 0` everywhere,
hence we are actually only solving the simple implicit diffusion equation
:math:`\kappa_\phi \nabla^2\phi = 0`,
which has exactly the uninteresting solution we obtained.

The resolution to this problem is to apply relaxation to obtain the desired
answer, i.e., the solution is allowed to relax in time from the initial
condition to the desired equilibrium solution.  To do so, we reintroduce the
transient term from Equation :eq:`eq-phase:simple`

.. index::
   single: TransientTerm

>>> eq = TransientTerm() == DiffusionTerm(coeff=kappa) + S0

>>> phase.setValue(1.)
>>> phase.setValue(0., where=x > L/2)

>>> for i in range(13):
...     eq.solve(var = phase, dt=1.)
...     if __name__ == '__main__':
...         viewer.plot()

After 13 time steps, the solution has converged to the analytical solution

>>> print(phase.allclose(analyticalArray, rtol = 1e-4, atol = 1e-4))
1
>>> from fipy import input
>>> if __name__ == '__main__':
...     input("Relaxation, explicit. Press <return> to proceed...")

.. image:: /figures/examples/phase/simple/relaxation.*
   :width: 50%
   :align: center
   :alt: steady-state equilibrium phase field after relaxation

.. note:: The solution is only found accurate
   to :math:`\approx 4.3\times 10^{-5}`
   because the infinite-domain analytical
   solution :eq:`eq-phase:simple:analytical`
   is not an exact representation for the solution in a finite domain of
   length :math:`L`.

Setting fixed-value boundary conditions of 1 and 0 would still require the
relaxation method with the fully explicit source.

Solution performance can be improved if we exploit the dependence of the
source on :math:`\phi`.  By doing so, we can make the source semi-implicit,
improving the rate of convergence over the fully explicit approach.  The
source can only be semi-implicit because we employ sparse linear algebra
routines to solve the PDEs, i.e., there is no fully implicit way to
represent a term like
:math:`\phi^4` in the linear set of equations
:math:`\mathsf{M} \vec{\phi} - \vec{b} = 0`.

By linearizing a source as
:math:`S = S_0 - S_1 \phi`,
we make it more implicit by adding the coefficient
:math:`S_1`
to the matrix diagonal.  For numerical stability, this linear coefficient
must never be negative.

There are an infinite number of choices for this linearization, but
many do not converge very well. One choice is that used by Ryo
Kobayashi:

.. index::
   single: ImplicitSourceTerm

>>> S0 = mPhi * phase * (mPhi > 0)
>>> S1 = mPhi * ((mPhi < 0) - phase)
>>> eq = DiffusionTerm(coeff=kappa) + S0 \
...   + ImplicitSourceTerm(coeff = S1)

.. note:: Because ``mPhi`` is a variable field, the quantities ``(mPhi > 0)``
   and ``(mPhi < 0)`` evaluate to variable *fields* of `True` and `False`,
   instead of single Boolean values.

This expression converges to the same value given by the explicit
relaxation approach, but in only 8 sweeps (note that because there is no
transient term, these sweeps are not time steps, but rather repeated
iterations at the same time step to reach a converged solution).

.. note::

   We use :meth:`~fipy.terms.term.Term.solve` instead of
   :meth:`~fipy.terms.term.Term.sweep` because we don't care about the residual.
   Either function would work, but :meth:`~fipy.terms.term.Term.solve` is a bit faster.

>>> phase.setValue(1.)
>>> phase.setValue(0., where=x > L/2)

>>> for i in range(8):
...     eq.solve(var = phase)
>>> print(phase.allclose(analyticalArray, rtol = 1e-4, atol = 1e-4))
1
>>> from fipy import input
>>> if __name__ == '__main__':
...     viewer.plot()
...     input("Kobayashi, semi-implicit. Press <return> to proceed...")

In general, the best convergence is obtained when the linearization gives a
good representation of the relationship between the source and the
dependent variable.  The best practical advice is to perform a Taylor
expansion of the source about the previous value of the dependent variable
such that
:math:`S = S_\text{old} + \left.\frac{\partial S}{\partial
\phi}\right\rvert_\text{old} (\phi - \phi_\text{old}) = (S -
\frac{\partial S}{\partial\phi} \phi)_\text{old} +
\left.\frac{\partial S}{\partial \phi}\right|_\text{old} \phi`.
Now, if our source term is represented by :math:`S = S_0 + S_1 \phi`,
then :math:`S_1 = \left.\frac{\partial S}{\partial
\phi}\right|_\text{old}` and :math:`S_0 = (S - \frac{\partial
S}{\partial\phi} \phi)_\text{old} = S_\text{old} - S_1
\phi_\text{old}`.
In this way, the linearized source will be tangent to the curve of the
actual source as a function of the dependent variable.

For our source,
:math:`S = m_\phi \phi (1 - \phi)`,

.. math::

   \frac{\partial S}{\partial \phi}
   = \frac{\partial m_\phi}{\partial \phi} \phi (1 - \phi) + m_\phi (1 - 2\phi)

and

.. math::

   \frac{\partial m_\phi}{\partial \phi} = 2 W - 30 (1 - 2\phi) L\frac{T-T_M}{T_M},

or

>>> dmPhidPhi = 2 * W - 30 * (1 - 2 * phase) * enthalpy
>>> S1 = dmPhidPhi * phase * (1 - phase) + mPhi * (1 - 2 * phase)
>>> S0 = mPhi * phase * (1 - phase) - S1 * phase
>>> eq = DiffusionTerm(coeff=kappa) + S0 \
...   + ImplicitSourceTerm(coeff = S1)

Using this scheme, where the coefficient of the implicit source term is
tangent to the source, we reach convergence in only 5 sweeps

>>> phase.setValue(1.)
>>> phase.setValue(0., where=x > L/2)

>>> for i in range(5):
...     eq.solve(var = phase)
>>> print(phase.allclose(analyticalArray, rtol = 1e-4, atol = 1e-4))
1
>>> from fipy import input
>>> if __name__ == '__main__':
...     viewer.plot()
...     input("Tangent, semi-implicit. Press <return> to proceed...")

Although, for this simple problem, there is no appreciable difference in
run-time between the fully explicit source and the optimized semi-implicit
source, the benefit of 60% fewer sweeps should be obvious for larger
systems and longer iterations.

----

This example has focused on just the region of the phase field interface in
equilibrium. Problems of interest, though, usually involve the dynamics of one phase
transforming to another. To that end, let us recast the problem using
physical parameters and dimensions. We'll need a new mesh

>>> nx = 400
>>> dx = 5e-6 # cm
>>> L = nx * dx

>>> mesh = Grid1D(dx = dx, nx = nx)

and thus must redeclare :math:`\phi` on the new mesh

>>> phase = CellVariable(name="phase",
...                      mesh=mesh,
...                      hasOld=1)
>>> x = mesh.cellCenters[0]
>>> phase.setValue(1.)
>>> phase.setValue(0., where=x > L/2)

We choose the parameter values appropriate for nickel, given in
:cite:`Warren:1995`

.. index::
   single: Variable

>>> Lv = 2350 # J / cm**3
>>> Tm = 1728. # K
>>> T = Variable(value=Tm)
>>> enthalpy = Lv * (T - Tm) / Tm # J / cm**3

The parameters of the phase field model can be related to the surface
energy :math:`\sigma` and the interfacial thickness :math:`\delta` by

.. math::

   \kappa &= 6\sigma\delta \\
   W &= \frac{6\sigma}{\delta} \\
   M_\phi &= \frac{T_m\beta}{6 L \delta}.

We take :math:`\delta \approx \Delta x`.

>>> delta = 1.5 * dx
>>> sigma = 3.7e-5 # J / cm**2
>>> beta = 0.33 # cm / (K s)
>>> kappa = 6 * sigma * delta # J / cm
>>> W = 6 * sigma / delta # J / cm**3
>>> Mphi = Tm * beta / (6. * Lv * delta) # cm**3 / (J s)

>>> if __name__ == '__main__':
...     displacement = L * 0.1
... else:
...     displacement = L * 0.025

>>> analyticalArray = CellVariable(name="tanh", mesh=mesh,
...                                value=0.5 * (1 - numerix.tanh((x - (L / 2. + displacement))
...                                                              / (2 * delta))))

and make a new viewer

>>> if __name__ == '__main__':
...     viewer2 = Viewer(vars = (phase, analyticalArray))
...     viewer2.plot()

Now we can redefine the transient phase field equation, using the optimal
form of the source term shown above

>>> mPhi = -((1 - 2 * phase) * W + 30 * phase * (1 - phase) * enthalpy)
>>> dmPhidPhi = 2 * W - 30 * (1 - 2 * phase) * enthalpy
>>> S1 = dmPhidPhi * phase * (1 - phase) + mPhi * (1 - 2 * phase)
>>> S0 = mPhi * phase * (1 - phase) - S1 * phase
>>> eq = TransientTerm(coeff=1/Mphi) == DiffusionTerm(coeff=kappa) \
...                         + S0 + ImplicitSourceTerm(coeff = S1)

In order to separate the effect of forming the phase field interface
from the kinetics of moving it, we first equilibrate at the melting
point. We now use the :meth:`~fipy.terms.term.Term.sweep` method instead of
:meth:`~fipy.terms.term.Term.solve` because we require the residual.

.. index::
   single: sweep

>>> timeStep = 1e-6
>>> for i in range(10):
...     phase.updateOld()
...     res = 1e+10
...     while res > 1e-5:
...         res = eq.sweep(var=phase, dt=timeStep)
>>> if __name__ == '__main__':
...     viewer2.plot()

and then quench by 1 K

>>> T.setValue(T() - 1)

In order to have a stable numerical solution, the interface must not move
more than one grid point per time step,
we thus set the timestep according to the grid spacing :math:`\Delta x`,
the linear kinetic coefficient :math:`\beta`, and the undercooling
:math:`\abs{T_m - T}`
Again we use the :meth:`~fipy.terms.term.Term.sweep` method as a replacement for
:meth:`~fipy.terms.term.Term.solve`.

>>> velocity = beta * abs(Tm - T()) # cm / s
>>> timeStep = .1 * dx / velocity # s
>>> elapsed = 0
>>> while elapsed < displacement / velocity:
...     phase.updateOld()
...     res = 1e+10
...     while res > 1e-5:
...         res = eq.sweep(var=phase, dt=timeStep)
...     elapsed += timeStep
...     if __name__ == '__main__':
...         viewer2.plot()

A hyperbolic tangent is not an exact steady-state solution given the
quintic polynomial we chose for the :math:`p` function, but it gives a
reasonable approximation.

>>> print(phase.allclose(analyticalArray, rtol = 5, atol = 2e-3))
1

If we had made another common choice of :math:`p(\phi) = \phi^2(3 - 2\phi)`,
we would have found much better agreement, as that case does give an
exact :math:`\tanh` solution in steady state.
If SciPy is available, another way to compare against the expected result
is to do a least-squared fit to determine the interface velocity and
thickness

.. index::
   pair: module; scipy

>>> try:
...     def tanhResiduals(p, y, x, t):
...         V, d = p
...         return y - 0.5 * (1 - numerix.tanh((x - V * t - L / 2.) / (2*d)))
...     from scipy.optimize import leastsq # doctest: +SCIPY
...     x =  mesh.cellCenters[0]
...     (V_fit, d_fit), msg = leastsq(tanhResiduals, [L/2., delta],
...                                   args=(phase.globalValue, x.globalValue, elapsed)) # doctest: +SCIPY
... except ImportError:
...     V_fit = d_fit = 0
...     print("The SciPy library is unavailable to fit the interface \
... thickness and velocity")

>>> print(abs(1 - V_fit / velocity) < 4.2e-2) # doctest: +SCIPY
True
>>> print(abs(1 - d_fit / delta) < 2e-2) # doctest: +SCIPY
True

>>> from fipy import input
>>> if __name__ == '__main__':
...     input("Dimensional, semi-implicit. Press <return> to proceed...")

.. image:: /figures/examples/phase/simple/dimensional.*
   :width: 90%
   :align: center
   :alt: phase field when solved with physical parameters

.. .. bibmissing:: /refs.bib
    :sort:
"""
__docformat__ = 'restructuredtext'

from fipy import input

if __name__ == '__main__':
    import fipy.tests.doctestPlus
    exec(fipy.tests.doctestPlus._getScript())

    input('finished')



__docformat__ = 'restructuredtext'


from fipy.terms.implicitSourceTerm import ImplicitSourceTerm
from fipy.variables.levelSetDiffusionVariable import _LevelSetDiffusionVariable
from fipy.terms.transientTerm import TransientTerm
from fipy.terms.diffusionTerm import DiffusionTermNoCorrection

def buildMetalIonDiffusionEquation(ionVar = None,
                                   distanceVar = None,
                                   depositionRate = 1,
                                   transientCoeff = 1,
                                   diffusionCoeff = 1,
                                   metalIonMolarVolume = 1):

    r"""

    The `MetalIonDiffusionEquation` solves the diffusion of the metal
    species with a source term at the electrolyte interface. The governing
    equation is given by,

    .. math::

       \frac{\partial c}{\partial t} = \nabla \cdot D \nabla  c

    where,

    .. math::

       D = \begin{cases}
           D_c & \text{when $\phi > 0$} \\
           0  & \text{when $\phi \le 0$}
       \end{cases}

    The velocity of the interface generally has a linear dependence on ion
    concentration. The following boundary condition applies at the zero
    level set,

    .. math::

       D \hat{n} \cdot \nabla c = \frac{v(c)}{\Omega} \qquad \text{at $phi = 0$}

    where

    .. math::

       v(c) = c V_0

    The test case below is for a 1D steady state problem. The solution is
    given by:

    .. math::

       c(x) = \frac{c^{\infty}}{\Omega D / V_0 + L}\left(x - L\right) + c^{\infty}

    This is the test case,

    >>> from fipy.meshes import Grid1D
    >>> nx = 11
    >>> dx = 1.
    >>> from fipy.tools import serialComm
    >>> mesh = Grid1D(nx = nx, dx = dx, communicator=serialComm)
    >>> x, = mesh.cellCenters
    >>> from fipy.variables.cellVariable import CellVariable
    >>> ionVar = CellVariable(mesh = mesh, value = 1.)
    >>> from fipy.variables.distanceVariable \
    ...     import DistanceVariable
    >>> disVar = DistanceVariable(mesh = mesh,
    ...                           value = (x - 0.5) - 0.99,
    ...                           hasOld = 1)

    >>> v = 1.
    >>> diffusion = 1.
    >>> omega = 1.
    >>> cinf = 1.

    >>> eqn = buildMetalIonDiffusionEquation(ionVar = ionVar,
    ...                                      distanceVar = disVar,
    ...                                      depositionRate = v * ionVar,
    ...                                      diffusionCoeff = diffusion,
    ...                                      metalIonMolarVolume = omega)

    >>> ionVar.constrain(cinf, mesh.facesRight)

    >>> for i in range(10):
    ...     eqn.solve(ionVar, dt = 1000)

    >>> L = (nx - 1) * dx - dx / 2
    >>> gradient = cinf / (omega * diffusion / v + L)
    >>> answer = gradient * (x - L - dx * 3 / 2) + cinf
    >>> answer[x < dx] = 1
    >>> print(ionVar.allclose(answer))
    1

    Testing the interface source term

    >>> from fipy.meshes import Grid2D
    >>> from fipy import numerix, serialComm
    >>> mesh = Grid2D(dx = 1., dy = 1., nx = 2, ny = 2, communicator=serialComm)
    >>> from fipy.variables.distanceVariable import DistanceVariable
    >>> distance = DistanceVariable(mesh = mesh, value = (-.5, .5, .5, 1.5))
    >>> ionVar = CellVariable(mesh = mesh, value = (1, 1, 1, 1))
    >>> depositionRate = CellVariable(mesh=mesh, value=(1, 1, 1, 1))
    >>> source = depositionRate * distance.cellInterfaceAreas / mesh.cellVolumes / ionVar
    >>> sqrt = numerix.sqrt(2)
    >>> ans = CellVariable(mesh=mesh, value=(0, 1 / sqrt, 1 / sqrt, 0))
    >>> print(numerix.allclose(source, ans))
    True
    >>> distance[:] = (-1.5, -0.5, -0.5, 0.5)
    >>> print(numerix.allclose(source, (0, 0, 0, sqrt)))
    True

    Parameters
    ----------
    ionVar : ~fipy.variables.cellVariable.CellVariable
         The metal ion concentration variable.
    distanceVar : ~fipy.variables.distanceVariable.DistanceVariable
    depositionRate : float or ~fipy.variables.cellVariable.CellVariable
         Represents the interface deposition rate.
    transientCoeff : float
         The transient coefficient.
    diffusionCoeff : float or ~fipy.variables.faceVariable.FaceVariable
         The diffusion coefficient
    metalIonMolarVolume : float
         Molar volume of the metal ions.

    """

    diffusionCoeff = _LevelSetDiffusionVariable(distanceVar,
                                                diffusionCoeff)

    eq =  TransientTerm(transientCoeff) - DiffusionTermNoCorrection(diffusionCoeff)

    mesh = distanceVar.mesh
    coeff = depositionRate * distanceVar.cellInterfaceAreas / (mesh.cellVolumes * metalIonMolarVolume) / ionVar

    return eq + ImplicitSourceTerm(coeff)

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()




from fipy.solvers.preconditioner import SolverModifyingPreconditioner

__all__ = ["AMGPreconditioner",
           "AggregationAMGPreconditioner",
           "BiCGStabPreconditioner",
           "CGPreconditioner",
           "DILUPreconditioner",
           "FGMRESPreconditioner",
           "GaussSeidelPreconditioner",
           "ILUPreconditioner",
           "JacobiPreconditioner",
           "PolynomialPreconditioner",
           "PyAMGXPreconditioner"]

class PyAMGXPreconditioner(SolverModifyingPreconditioner):
    """Interface to pyamgx_ `preconditioner configuration`_ for :class:`~fipy.solvers.pyamgx.pyAMGXSolver.PyAMGXSolver`.

    .. _pyamgx: https://pyamgx.readthedocs.io
    .. _preconditioner configuration: https://pyamgx.readthedocs.io/en/latest/basic.html#config-objects
    """

    def __init__(self, **kwargs):
        """
        Parameters
        ----------
        **kwargs : dict, optional
            Extra arguments to preconditioner: refer to `preconditioner
            configuration`_ for information about possible arguments.
        """
        self.config_dict = {
            "solver": self.pctype,
            "max_iters": 1
        }
        self.config_dict.update(kwargs)

    def _applyToSolver(self, solver, matrix=None):
        solver["preconditioner"] = self.config_dict.copy()

class AMGPreconditioner(PyAMGXPreconditioner):
    """Adaptive Multigrid preconditioner for :class:`~fipy.solvers.pyamgx.pyAMGXSolver.PyAMGXSolver`.
    """

    pctype = "AMG"

class AggregationAMGPreconditioner(AMGPreconditioner):
    """Aggregation Adaptive Multigrid preconditioner for :class:`~fipy.solvers.pyamgx.pyAMGXSolver.PyAMGXSolver`.
    """

    def __init__(self):
        super(ClassicalAMGPreconditioner, self).__init__(algorithm="AGGREGATION",
                                                         selector="SIZE_2")
class BiCGStabPreconditioner(PyAMGXPreconditioner):
    """Biconjugate Gradient Stabilized preconditioner for :class:`~fipy.solvers.pyamgx.pyAMGXSolver.PyAMGXSolver`.
    """

    pctype = "PCIBCGSTAB"

class CGPreconditioner(PyAMGXPreconditioner):
    """Conjugate Gradient preconditioner for :class:`~fipy.solvers.pyamgx.pyAMGXSolver.PyAMGXSolver`.
    """

    pctype = "PCG"

class DILUPreconditioner(PyAMGXPreconditioner):
    """Diagonal Incomplete LU preconditioner for :class:`~fipy.solvers.pyamgx.pyAMGXSolver.PyAMGXSolver`.
    """

    pctype = "MULTICOLOR_DILU"

class FGMRESPreconditioner(PyAMGXPreconditioner):
    """Flexible Generalized Minimum Residual preconditioner for :class:`~fipy.solvers.pyamgx.pyAMGXSolver.PyAMGXSolver`.
    """

    pctype = "FGMRES"

class GaussSeidelPreconditioner(PyAMGXPreconditioner):
    """Gauss-Seidel preconditioner for :class:`~fipy.solvers.pyamgx.pyAMGXSolver.PyAMGXSolver`.
    """

    pctype = "MULTICOLOR_GS"

class ILUPreconditioner(PyAMGXPreconditioner):
    """Incomplete LU preconditioner for :class:`~fipy.solvers.pyamgx.pyAMGXSolver.PyAMGXSolver`.
    """

    pctype = "MULTICOLOR_GS"

class JacobiPreconditioner(PyAMGXPreconditioner):
    """Block Jacobi preconditioner for :class:`~fipy.solvers.pyamgx.pyAMGXSolver.PyAMGXSolver`.
    """

    pctype = "BLOCK_JACOBI"

class PolynomialPreconditioner(PyAMGXPreconditioner):
    """Polynomial preconditioner  for :class:`~fipy.solvers.pyamgx.pyAMGXSolver.PyAMGXSolver`.
    """

    pctype = "POLYNOMIAL"



__docformat__ = 'restructuredtext'

from PyTrilinos import AztecOO

from fipy.solvers.trilinos.trilinosAztecOOSolver import TrilinosAztecOOSolver
from fipy.solvers.trilinos.preconditioners.multilevelDDPreconditioner import MultilevelDDPreconditioner

__all__ = ["LinearCGSolver", "LinearPCGSolver"]

class LinearCGSolver(TrilinosAztecOOSolver):

    """Interface to the conjugate gradient (:term:`CG`) solver in
    :ref:`TRILINOS`.

    Uses the
    :class:`~fipy.solvers.trilinos.preconditioners.multilevelDDPreconditioner.MultilevelDDPreconditioner`
    by default.
    """

    solver = AztecOO.AZ_cg

    DEFAULT_PRECONDITIONER = MultilevelDDPreconditioner

    def _canSolveAsymmetric(self):
        return False

LinearPCGSolver = LinearCGSolver



r"""Solve the Cahn-Hilliard problem in two dimensions.

.. warning:: This formulation has
   `serious performance problems`_
   and is **not automatically tested**.  Specifically, for non-trivial mesh
   sizes, (obsolete) Pysparse required enormous amounts of memory,
   :ref:`Trilinos` cannot solve the coupled form, and :ref:`PETSc` cannot
   solve the vector form.

.. _serious performance problems: https://github.com/usnistgov/fipy/issues/378

The spinodal decomposition phenomenon is a spontaneous separation of
an initially homogeneous mixture into two distinct regions of different
properties (spin-up/spin-down, component A/component B). It is a
"barrierless" phase separation process, such that under the right
thermodynamic conditions, any fluctuation, no matter how small, will
tend to grow. This is in contrast to nucleation, where a fluctuation
must exceed some critical magnitude before it will survive and grow.
Spinodal decomposition can be described by the "Cahn-Hilliard"
equation (also known as
"conserved Ginsberg-Landau" or "model B" of Hohenberg & Halperin)

.. math::

   \frac{\partial \phi}{\partial t}
   = \nabla \cdot D \nabla \left( \frac{\partial f}{\partial \phi}   - \epsilon^2 \nabla^2 \phi \right).

where :math:`\phi` is a conserved order parameter, possibly representing
alloy composition or spin.
The double-well free energy function :math:`f = (a^2/2) \phi^2 (1 - \phi)^2`
penalizes states with intermediate values of :math:`\phi`
between 0 and 1. The gradient energy term :math:`\epsilon^2 \nabla^2\phi`,
on the other hand, penalizes sharp changes of :math:`\phi`.
These two competing effects result in the segregation
of :math:`\phi` into domains of 0 and 1, separated by abrupt, but
smooth, transitions. The parameters :math:`a` and :math:`\epsilon` determine the relative
weighting of the two effects and :math:`D` is a rate constant.

We can simulate this process in :term:`FiPy` with a simple script:

>>> from fipy import CellVariable, Grid2D, GaussianNoiseVariable, DiffusionTerm, TransientTerm, ImplicitSourceTerm, Viewer
>>> from fipy.tools import numerix

(Note that all of the functionality of NumPy is imported along with :term:`FiPy`, although
much is augmented for :term:`FiPy`\'s needs.)

>>> if __name__ == "__main__":
...     nx = ny = 20
... else:
...     nx = ny = 10
>>> mesh = Grid2D(nx=nx, ny=ny, dx=0.25, dy=0.25)
>>> phi = CellVariable(name=r"$\phi$", mesh=mesh)
>>> psi = CellVariable(name=r"$\psi$", mesh=mesh)

We start the problem with random fluctuations about :math:`\phi = 1/2`

>>> noise = GaussianNoiseVariable(mesh=mesh,
...                               mean=0.5,
...                               variance=0.01).value

>>> phi[:] = noise

:term:`FiPy` doesn't plot or output anything unless you tell it to:

>>> if __name__ == "__main__":
...     viewer = Viewer(vars=(phi,), datamin=0., datamax=1.)

We factor the Cahn-Hilliard equation into two 2nd-order PDEs and place
them in canonical form for :term:`FiPy` to solve them as a coupled set
of equations.

.. math::

   \frac{\partial \phi}{\partial t} &= \nabla\cdot D \nabla \psi \\
   \psi &= \left(\frac{\partial f}{\partial \phi} - \frac{\partial^2 f}{\partial \phi^2}\phi\right)_{\text{old}}
           + \frac{\partial^2 f}{\partial \phi^2}\phi - \epsilon^2 \nabla^2 \phi

The source term in :math:`\psi`, :math:`\frac{\partial f}{\partial \phi}`, is
expressed in linearized form after Taylor expansion at :math:`\phi=\phi_{\text{old}}`,
for the same reasons discussed in :mod:`examples.phase.simple`.
We need to perform the partial derivatives

.. math::

   \frac{\partial f}{\partial \phi} &= (a^2/2) 2 \phi (1 - \phi) (1 - 2 \phi) \\
   \frac{\partial^2 f}{\partial \phi^2} &= (a^2/2) 2 \left[1 - 6 \phi(1 - \phi)\right]

manually.

>>> D = a = epsilon = 1.
>>> dfdphi = a**2 * phi * (1 - phi) * (1 - 2 * phi)
>>> dfdphi_ = a**2 * (1 - phi) * (1 - 2 * phi)
>>> d2fdphi2 = a**2 * (1 - 6 * phi * (1 - phi))
>>> eq1 = (TransientTerm(var=phi) == DiffusionTerm(coeff=D, var=psi))
>>> eq2 = (ImplicitSourceTerm(coeff=1., var=psi)
...        == ImplicitSourceTerm(coeff=d2fdphi2, var=phi) - d2fdphi2 * phi + dfdphi
...        - DiffusionTerm(coeff=epsilon**2, var=phi))
>>> eq3 = (ImplicitSourceTerm(coeff=1., var=psi)
...        == ImplicitSourceTerm(coeff=dfdphi_, var=phi)
...        - DiffusionTerm(coeff=epsilon**2, var=phi))

>>> eq = eq1 & eq2

Because the evolution of a spinodal microstructure slows with time, we
use exponentially increasing time steps to keep the simulation
"interesting". The :term:`FiPy` user always has direct control over the
evolution of their problem.

>>> dexp = -5
>>> elapsed = 0.
>>> if __name__ == "__main__":
...     duration = 100.
... else:
...     duration = .5e-1

>>> while elapsed < duration:
...     dt = min(100, numerix.exp(dexp))
...     elapsed += dt
...     dexp += 0.01
...     eq.solve(dt=dt)
...     if __name__ == "__main__":
...         viewer.plot()

>>> from fipy import input
>>> if __name__ == '__main__':
...     input("Coupled equations. Press <return> to proceed...")

.. image:: /figures/examples/cahnHilliard/mesh2DCoupled.*
   :width: 90%
   :align: center

----

These equations can also be solved in :term:`FiPy` using a vector
equation. The variables :math:`\phi` and :math:`\psi` are now stored in
a single variable

>>> var = CellVariable(mesh=mesh, elementshape=(2,))
>>> var[0] = noise

>>> if __name__ == "__main__":
...     viewer = Viewer(name=r"$\phi$", vars=var[0,], datamin=0., datamax=1.)

>>> D = a = epsilon = 1.
>>> v0 = var[0]
>>> dfdphi = a**2 * v0 * (1 - v0) * (1 - 2 * v0)
>>> dfdphi_ = a**2 * (1 - v0) * (1 - 2 * v0)
>>> d2fdphi2 = a**2 * (1 - 6 * v0 * (1 - v0))

The source terms have to be shaped correctly for a vector. The implicit source
coefficient has to have a shape of `(2, 2)` while the explicit source
has a shape `(2,)`

>>> source = (- d2fdphi2 * v0 + dfdphi) * (0, 1)
>>> impCoeff = d2fdphi2 * ((0, 0),
...                        (1., 0)) + ((0, 0),
...                                    (0, -1.))

This is the same equation as the previous definition of `eq`, but now in
a vector format.

>>> eq = (TransientTerm(((1., 0.),
...                     (0., 0.))) == DiffusionTerm([((0.,          D),
...                                                   (-epsilon**2, 0.))])
...                                   + ImplicitSourceTerm(impCoeff) + source)

>>> dexp = -5
>>> elapsed = 0.

>>> while elapsed < duration:
...     dt = min(100, numerix.exp(dexp))
...     elapsed += dt
...     dexp += 0.01
...     eq.solve(var=var, dt=dt)
...     if __name__ == "__main__":
...         viewer.plot()

>>> print(numerix.allclose(var, (phi, psi)))
True

"""
__docformat__ = 'restructuredtext'

from fipy import input

if __name__ == '__main__':
    import fipy.tests.doctestPlus
    exec(fipy.tests.doctestPlus._getScript())

    input('finished')



"""Solving sparse linear systems
"""
__docformat__ = 'restructuredtext'


import logging

_log = logging.getLogger(__name__)

import os
from importlib import import_module

from fipy.tools.parser import _parseSolver

from fipy.solvers.solver import *

_desired_solver = _parseSolver()

if _desired_solver is None and 'FIPY_SOLVERS' in os.environ:
    _desired_solver = os.environ['FIPY_SOLVERS'].lower()
del os
    
try:
    from mpi4py import MPI
    _Nproc = MPI.COMM_WORLD.size
    del MPI
except ImportError:
    _Nproc = 1

_exceptions = []

class SerialSolverError(Exception):
    def __init__(self):
        super(SerialSolverError, self).__init__('solver does not run in parallel')

_exceptions = {}

def _import_matrix(suite, kind):
    """`from fipy.matrices.suiteMatrix import _SuiteKindMatrix`
    """
    m = import_module("fipy.matrices.{}Matrix".format(suite.lower()))
    return getattr(m, "_{}{}Matrix".format(suite, kind))

def _import_mesh_matrices(suite):
    """Import row, column, and general mesh matrices from `suite`
    """
    _RowMeshMatrix = _import_matrix(suite=suite, kind="RowMesh")
    _ColMeshMatrix = _import_matrix(suite=suite, kind="ColMesh")
    _MeshMatrix = _import_matrix(suite=suite, kind="Mesh")

    return _RowMeshMatrix, _ColMeshMatrix, _MeshMatrix

solver_suite = None

# The following definitions are only to provide documentation.
# They will be overridden by the solver suite that's actually imported.
DefaultSolver = None
"""Solver class for solving symmetric matrices.

This solver should be both robust and performant.
"""
DefaultAsymmetricSolver = None
"""Solver class for solving asymmetric matrices.
"""
DummySolver = None
"""Solver used by tests that don't actually need to solve.

Some tests are intended to confirm the matrix building machinery, but don't
actually need to solve (and may not be able to, e.g., zeros on the
diagonal).
"""
GeneralSolver = None
"""Solver class that should solve any matrix.
"""

from fipy.tools.comms.dummyComm import DummyComm
serialComm, parallelComm = DummyComm(), DummyComm()

if solver_suite is None and _desired_solver in ["petsc", None]:
    try:
        import petsc4py
        petsc4py.init()

        from fipy.solvers.petsc import *

        from fipy.solvers.petsc.comms.serialPETScCommWrapper import SerialPETScCommWrapper
        serialComm = SerialPETScCommWrapper()

        if _Nproc > 1:
            from fipy.solvers.petsc.comms.parallelPETScCommWrapper import ParallelPETScCommWrapper
            parallelComm = ParallelPETScCommWrapper()
        else:
            parallelComm = SerialPETScCommWrapper()

        _mesh_matrices = _import_mesh_matrices(suite="PETSc")
        solver_suite = "petsc"
    except Exception as inst:
        _exceptions["petsc"] = inst

if solver_suite is None and _desired_solver in ["trilinos", None]:
    try:
        from fipy.solvers.trilinos import *
        
        from fipy.solvers.trilinos.comms.serialEpetraCommWrapper import SerialEpetraCommWrapper
        serialComm = SerialEpetraCommWrapper()

        if _Nproc > 1:
            from fipy.solvers.trilinos.comms.parallelEpetraCommWrapper import ParallelEpetraCommWrapper
            parallelComm = ParallelEpetraCommWrapper()
        else:
            parallelComm = SerialEpetraCommWrapper()

        _mesh_matrices = _import_mesh_matrices(suite="Trilinos")
        solver_suite = "trilinos"
    except Exception as inst:
        _exceptions["trilinos"] = inst

if solver_suite is None and _desired_solver in ["scipy", None]:
    try:
        if _Nproc > 1:
            raise SerialSolverError()
        from fipy.solvers.scipy import *
        _mesh_matrices = _import_mesh_matrices(suite="Scipy")
        solver_suite = "scipy"

        # if PyAMG is installed, then fetch its preconditioner(s)
        try:
            from fipy.solvers.pyAMG.preconditioners import *
            _log.info("PyAMG preconditioners installed")
        except ImportError as imp:
            _log.info("PyAMG preconditioners not available: %s" % imp)
    except Exception as inst:
        _exceptions["scipy"] = inst

if solver_suite is None and _desired_solver in ["pyamgx", None]:
    try:
        if _Nproc > 1:
            raise  SerialSolverError('pyamgx')
        from fipy.solvers.pyamgx import *
        _mesh_matrices = _import_mesh_matrices(suite="Scipy")
        solver_suite = "pyamgx"
    except Exception as inst:
        _exceptions["pyamgx"] = inst

if solver_suite is None:
    if _desired_solver is None:
        raise ImportError('Unable to load a solver: %s' % str(_exceptions))
    else:
        if len(_exceptions) > 0:
            raise ImportError('Unable to load solver %s: %s' % (_desired_solver, _exceptions[_desired_solver]))
        else:
            raise ImportError('Unknown solver package %s' % _desired_solver)

# don't unpack until here in order to keep code above more succinct
_RowMeshMatrix, _ColMeshMatrix, _MeshMatrix = _mesh_matrices

INDEX_TYPE = _MeshMatrix.INDEX_TYPE

from fipy.tests.doctestPlus import register_skipper

register_skipper(flag='PETSC_SOLVER',
                 test=lambda: solver_suite == 'petsc',
                 why="the PETSc solvers are not being used.",
                 skipWarning=True)

register_skipper(flag='NOT_PYAMGX_SOLVER',
                 test=lambda: solver_suite != 'pyamgx',
                 why="the PyAMGX solver is being used.",
                 skipWarning=True)

register_skipper(flag='SCIPY_SOLVER',
                 test=lambda: solver_suite == 'scipy',
                 why="the SciPy solvers are not being used.",
                 skipWarning=True)

register_skipper(flag='PYAMGX_SOLVER',
                 test=lambda: solver_suite == 'pyamgx',
                 why="the pyamgx solvers are not being used.",
                 skipWarning=True)

register_skipper(flag='TRILINOS_SOLVER',
                 test=lambda: solver_suite == 'trilinos',
                 why="the Trilinos solvers are not being used.",
                 skipWarning=True)
                 
register_skipper(flag='NOT_TRILINOS_SOLVER',
                 test=lambda: solver_suite != 'trilinos',
                 why="the Trilinos solvers are being used.",
                 skipWarning=True)

del register_skipper

_log.info("Solver suite is %s", solver_suite)
_log.debug("DefaultSolver is %s", DefaultSolver)
_log.debug("DefaultAsymmetricSolver is %s", DefaultAsymmetricSolver)
_log.debug("DummySolver is %s", DummySolver)
_log.debug("GeneralSolver is %s", GeneralSolver)
_log.debug("serialComm is %s", serialComm)
_log.debug("parallelComm is %s", parallelComm)



__docformat__ = 'restructuredtext'

from PyTrilinos import IFPACK

from .trilinosPreconditioner import TrilinosPreconditioner

__all__ = ["ICPreconditioner"]

class ICPreconditioner(TrilinosPreconditioner):
    """Incomplete Cholesky Preconditioner from IFPACK for :class:`~fipy.solvers.trilinos.trilinosSolver.TrilinosSolver`.
    """

    def _applyToSolver(self, solver, matrix):
        Factory = IFPACK.Factory()
        Prec = Factory.Create("IC", matrix)
        Prec.Initialize()
        Prec.Compute()
        solver.SetPrecOperator(Prec)



from fipy.tests.doctestPlus import _LateImportDocTestSuite
import fipy.tests.testProgram

def _suite():
    return _LateImportDocTestSuite(docTestModuleNames = (
            'howToWriteAScript',
            'simpleTrenchSystem',
            'gold',
            'leveler',
            'gapFillMesh',
            'trenchMesh',
            'metalIonDiffusionEquation',
            'lines',
            'adsorbingSurfactantEquation',
        ), base = __name__)

if __name__ == '__main__':
    fipy.tests.testProgram.main(defaultTest='_suite')



r"""The following examples exhibit various parts of a model to study
electrochemical interfaces.  In a pair of papers, Guyer, Boettinger, Warren
and McFadden :cite:`ElPhFI` :cite:`ElPhFII` have shown that an electrochemical
interface can be modeled by an equation for the phase field :math:`\xi`

.. math::
   :label: elphf:phase

    \underbrace{
	\frac{1}{M_\xi}\frac{\partial \xi}{\partial t}
	\vphantom{\sum_{j=1}^{n} C_j}
    }_{\text{transient}}
    = 
    \underbrace{
	\kappa_{\xi}\nabla^2 \xi
	\vphantom{\sum_{j=1}^{n} C_j}
    }_{\text{diffusion}}
    - 
    \underbrace{
	\overbrace{
	    \sum_{j=1}^{n} C_j \left[
		p'(\xi) \Delta\mu_j^\circ
		+ g'(\xi) W_j
	    \right]
	}^{\text{phase transformation}}
	+
	\overbrace{
	    \frac{\epsilon'(\xi)}{2}\left(\nabla\phi\right)^2
	    \vphantom{\sum_{j=1}^{n} C_j}
	}^{\text{dielectric}}
    }_{\text{source}}


a set of diffusion equations for the concentrations :math:`C_j`, for the substitutional elements
:math:`j = 2,\ldots, n-1`

.. math::
   :label: elphf:substitutional

   \underbrace{
	\frac{\partial C_j}{\partial t}
   }_{\text{transient}}
    &= \underbrace{
	D_{j}\nabla^2 C_j
	\vphantom{\frac{\partial C_j}{\partial t}}
    }_{\text{diffusion}} \\
    & \qquad + \underbrace{
	D_{j}\nabla\cdot 
	\frac{C_j}{1 - \sum_{\substack{k=2\\ k \neq j}}^{n-1} C_k}
	\left\{
	    \overbrace{
		\sum_{\substack{i=2\\ i \neq j}}^{n-1} \nabla C_i
	    }^{\text{counter diffusion}}
	   + 
	   \overbrace{ 
		C_n \left[
		    p'(\xi) \Delta\mu_{jn}^{\circ}
		    + g'(\xi) W_{jn}
		\right] \nabla\xi
		\vphantom{\sum_{\substack{i=2\\ i \neq j}}^{n-1} \nabla C_i}
	   }^{\text{phase transformation}}
	   +
	   \overbrace{
		C_n z_{jn} \nabla \phi
		\vphantom{\sum_{\substack{i=2\\ i \neq j}}^{n-1} \nabla C_i}
	   }^{\text{electromigration}}
	\right\}
    }_{\text{convection}}

a diffusion equation for the concentration :math:`C_{\text{e}^{-}}` of
electrons

.. math::
    :label: elphf:interstitial

    \underbrace{
	\frac{\partial C_{\text{e}^{-}}}{\partial t}
	\vphantom{\left\{
	    \overbrace{
		\left[p'(\xi)\right]
	    }^{\text{phase transformation}}
	\right\}}
    }_{\text{transient}}
    = \underbrace{
	D_{\text{e}^{-}}\nabla^2 C_{\text{e}^{-}} 
	\vphantom{\left\{
	    \overbrace{
		\left[p'(\xi)\right]
	    }^{\text{phase transformation}}
	\right\}}
    }_{\text{diffusion}} \\
    + \underbrace{
	D_{\text{e}^{-}}\nabla\cdot 
	C_{\text{e}^{-}}
	\left\{
	    \overbrace{
		\left[
		    p'(\xi) \Delta\mu_{\text{e}^{-}}^{\circ}
		    + g'(\xi) W_{\text{e}^{-}}
		\right] \nabla\xi
	    }^{\text{phase transformation}}
	    +
	    \overbrace{
		z_{\text{e}^{-}} \nabla \phi
		\vphantom{\left[p'(\xi)\right]}
	   }^{\text{electromigration}}
	\right\}
    }_{\text{convection}}

and Poisson's equation for the electrostatic potential :math:`\phi`

.. math::
    :label: elphf:Poisson

    \underbrace{
	\nabla\cdot\left(\epsilon\nabla\phi\right) 
	\vphantom{\sum_{j=1}^n z_j C_j}
    }_{\text{diffusion}}
    +
    \underbrace{
	\sum_{j=1}^n z_j C_j
    }_{\text{source}}
    = 0

:math:`M_\xi` is the phase field mobility, :math:`\kappa_\xi` is the phase
field gradient energy coefficient,
:math:`p'(\xi) = 30\xi^2\left(1-\xi\right)^2`, and
:math:`g'(\xi) = 2\xi\left(1-\xi\right)\left(1-2\xi\right)`.  For a given species
:math:`j`, :math:`\Delta\mu_j^{\circ}` is the standard chemical potential
difference between the electrode and electrolyte for a pure material,
:math:`W_j` is the magnitude of the energy barrier in the double-well
free energy function, :math:`z_j` is the valence, and :math:`D_{j}` is the
self diffusivity.  :math:`\Delta\mu_{jn}^{\circ}`, :math:`W_{jn}`, and
:math:`z_{jn}` are the differences of the respective quantities
:math:`\Delta\mu_{j}^{\circ}`, :math:`W_{j}`, and :math:`z_{j}` between
substitutional species :math:`j` and the solvent species :math:`n`.  The
total charge is denoted by :math:`\sum_{j=1}^n z_j C_j`.

Although unresolved stiffnesses make the full solution of this coupled
set of equations intractable in :term:`FiPy`, the following examples
demonstrate the setup and solution of various parts.
"""



__docformat__ = 'restructuredtext'

from fipy.solvers.preconditioner import MatrixModifyingPreconditioner

__all__ = ["ScipyPreconditioner"]

class ScipyPreconditioner(MatrixModifyingPreconditioner):
    """Base class for preconditioners for :class:`~fipy.solvers.scipy.scipySolver.ScipySolver`.

    .. attention:: This class is abstract. Always create one of its subclasses.
    """

    pass



# -*- coding: utf-8 -*-


r"""Solve the steady-state convection-diffusion equation in one dimension.

This example solves the steady-state convection-diffusion equation
given by

.. math::

   \nabla \cdot \left(D \nabla \phi + \vec{u} \phi \right) = 0

with coefficients :math:`D = 1` and :math:`\vec{u} = 10\hat{\imath}`, or

>>> diffCoeff = 1.
>>> convCoeff = (10.,)

We define a 1D mesh

.. index::
   single: Grid1D

>>> from fipy import CellVariable, Grid1D, DiffusionTerm, ExponentialConvectionTerm, Viewer
>>> from fipy.tools import numerix

>>> L = 10.
>>> nx = 10
>>> mesh = Grid1D(dx=L / nx, nx=nx)

>>> valueLeft = 0.
>>> valueRight = 1.

The solution variable is initialized to ``valueLeft``:

>>> var = CellVariable(mesh=mesh, name="variable")

and impose the boundary conditions

.. math::

   \phi = \begin{cases}
   0& \text{at $x = 0$,} \\
   1& \text{at $x = L$,}
   \end{cases}

with

>>> var.constrain(valueLeft, mesh.facesLeft)
>>> var.constrain(valueRight, mesh.facesRight)

The equation is created with the :class:`~fipy.terms.diffusionTerm.DiffusionTerm` and
:class:`~fipy.terms.exponentialConvectionTerm.ExponentialConvectionTerm`. The scheme used by the convection term
needs to calculate a Péclet number and thus the diffusion term
instance must be passed to the convection term.

>>> eq = (DiffusionTerm(coeff=diffCoeff)
...       + ExponentialConvectionTerm(coeff=convCoeff))

More details of the benefits and drawbacks of each type of convection
term can be found in :ref:`sec:NumericalSchemes`.
Essentially, the :class:`~fipy.terms.exponentialConvectionTerm.ExponentialConvectionTerm` and :class:`~fipy.terms.powerLawConvectionTerm.PowerLawConvectionTerm` will
both handle most types of convection-diffusion cases, with the
:class:`~fipy.terms.powerLawConvectionTerm.PowerLawConvectionTerm` being more efficient.

We solve the equation

>>> eq.solve(var=var)

and test the solution against the analytical result

.. math::

   \phi = \frac{1 - \exp(-u_x x / D)}{1 - \exp(-u_x L / D)}

or

.. index::
   single: exp

>>> axis = 0
>>> x = mesh.cellCenters[axis]
>>> CC = 1. - numerix.exp(-convCoeff[axis] * x / diffCoeff)
>>> DD = 1. - numerix.exp(-convCoeff[axis] * L / diffCoeff)
>>> analyticalArray = CC / DD
>>> print(var.allclose(analyticalArray))
1

If the problem is run interactively, we can view the result:

.. index::
   pair: module; fipy.viewers

>>> if __name__ == '__main__':
...     viewer = Viewer(vars=var)
...     viewer.plot()
"""
__docformat__ = 'restructuredtext'

from fipy import input

if __name__ == '__main__':
    import fipy.tests.doctestPlus
    exec(fipy.tests.doctestPlus._getScript())

    input('finished')




__docformat__ = 'restructuredtext'

from PyTrilinos import AztecOO

from .trilinosSolver import TrilinosSolver
from .preconditioners.jacobiPreconditioner import JacobiPreconditioner
from fipy.tools.timer import Timer

__all__ = ["TrilinosAztecOOSolver"]

class TrilinosAztecOOSolver(TrilinosSolver):

    """
    .. attention:: This class is abstract, always create on of its subclasses.
       It provides the code to call all solvers from the Trilinos AztecOO package.

    """

    DEFAULT_PRECONDITIONER = JacobiPreconditioner

    def _adaptLegacyTolerance(self, L, x, b):
        return self._adaptInitialTolerance(L, x, b)

    def _adaptUnscaledTolerance(self, L, x, b):
        return (1., AztecOO.AZ_noscaled)

    def _adaptRHSTolerance(self, L, x, b):
        return (1., AztecOO.AZ_rhs)

    def _adaptMatrixTolerance(self, L, x, b):
        return (1., AztecOO.AZ_Anorm)

    def _adaptInitialTolerance(self, L, x, b):
        return (1., AztecOO.AZ_r0)

    def _adaptSolutionTolerance(self, L, x, b):
        return (1., AztecOO.AZ_sol)

    def _solve_(self, L, x, b):
        """Solve system of equations posed for PyTrilinos

        Parameters
        ----------
        L : Epetra.CrsMatrix
            Sparse matrix
        x : Epetra.Vector
            Solution variable as non-ghosted vector
        b : Epetra.Vector
            Right-hand side as non-ghosted vector

        Returns
        -------
        x : Epetra.Vector
            Solution variable as non-ghosted vector
        """

        solver = AztecOO.AztecOO(L, x, b)
        solver.SetAztecOption(AztecOO.AZ_solver, self.solver)

##        solver.SetAztecOption(AztecOO.AZ_kspace, 30)

        solver.SetAztecOption(AztecOO.AZ_output, AztecOO.AZ_none)

        tolerance_scale, suite_criterion = self._adaptTolerance(L, x, b)

        rtol = self.scale_tolerance(self.tolerance, tolerance_scale)

        solver.SetAztecOption(AztecOO.AZ_conv, suite_criterion)

        self._log.debug("BEGIN precondition")

        with Timer() as t:
            if self.preconditioner is None:
                solver.SetAztecOption(AztecOO.AZ_precond, AztecOO.AZ_none)
            else:
                self.preconditioner._applyToSolver(solver=solver, matrix=L)

        self._log.debug("END precondition - {} ns".format(t.elapsed))

        self._log.debug("BEGIN solve")

        with Timer() as t:
            solver.Iterate(self.iterations, rtol)

        self._log.debug("END solve - {} ns".format(t.elapsed))

        if self.preconditioner is not None:
            if hasattr(self.preconditioner, 'Prec'):
                del self.preconditioner.Prec

        status = solver.GetAztecStatus()

        self._setConvergence(suite="trilinos",
                             code=int(status[AztecOO.AZ_why]),
                             iterations=int(status[AztecOO.AZ_its]),
                             tolerance_scale=tolerance_scale,
                             residual=status[AztecOO.AZ_r],
                             scaled_residual=status[AztecOO.AZ_scaled_r],
                             convergence_residual=status[AztecOO.AZ_rec_r],
                             solve_time=status[AztecOO.AZ_solve_time],
                             Aztec_version=status[AztecOO.AZ_Aztec_version])

        self.convergence.warn()

        return x



__docformat__ = 'restructuredtext'

__all__ = ["Preconditioner",
           "SolverModifyingPreconditioner",
           "MatrixModifyingPreconditioner"]

class Preconditioner(object):
    """Base class for solver preconditioners.

    .. attention:: This class is abstract. Always create one of its subclasses.
    """

    pass


class SolverModifyingPreconditioner(Preconditioner):
    """Base class for preconditioners that modify a :class:`~fipy.solvers.solver.Solver`.
    """

    def _applyToSolver(self, solver, matrix):
        """Modify `solver` to apply preconditioning to `matrix`.

        Parameters
        ----------
        solver
            The solver to modify with preconditioner.
        matrix
            The matrix the preconditioner applies to.

        Returns
        -------
        None
        """
        raise NotImplementedError


class MatrixModifyingPreconditioner(Preconditioner):
    """Base class for preconditioners that modify a :class:`~fipy.matrices.sparseMatrix._SparseMatrix`.
    """

    def _applyToMatrix(self, matrix):
        """Create a preconditioner for `matrix`.

        Returns
        -------
        preconditioner : object
            Preconditioning object appropriate for this solver suite.
        matrix : :class:`~fipy.matrices.sparseMatrix._SparseMatrix`
            Matrix, possibly restructured to facilitate applying
            preconditioner.
        """
        raise NotImplementedError




r"""
The spinodal decomposition phenomenon is a spontaneous separation of
an initially homogeneous mixture into two distinct regions of different
properties (spin-up/spin-down, component A/component B). It is a
"barrierless" phase separation process, such that under the right
thermodynamic conditions, any fluctuation, no matter how small, will
tend to grow. This is in contrast to nucleation, where a fluctuation
must exceed some critical magnitude before it will survive and grow.
Spinodal decomposition can be described by the "Cahn-Hilliard"
equation (also known as
"conserved Ginsberg-Landau" or "model B" of Hohenberg & Halperin)

.. math::

   \frac{\partial \phi}{\partial t}
   = \nabla\cdot D \nabla\left( \frac{\partial f}{\partial \phi}   - \epsilon^2 \nabla^2 \phi\right).

where :math:`\phi` is a conserved order parameter, possibly representing
alloy composition or spin.
The double-well free energy function :math:`f = (a^2/2) \phi^2 (1 - \phi)^2`
penalizes states with intermediate values of :math:`\phi`
between 0 and 1. The gradient energy term :math:`\epsilon^2 \nabla^2\phi`,
on the other hand, penalizes sharp changes of :math:`\phi`.
These two competing effects result in the segregation
of :math:`\phi` into domains of 0 and 1, separated by abrupt, but
smooth, transitions. The parameters :math:`a` and :math:`\epsilon` determine the relative
weighting of the two effects and :math:`D` is a rate constant.

We can simulate this process in :term:`FiPy` with a simple script:

>>> from fipy import CellVariable, Grid2D, GaussianNoiseVariable, TransientTerm, DiffusionTerm, ImplicitSourceTerm, LinearLUSolver, Viewer, DefaultSolver
>>> from fipy.tools import numerix

(Note that all of the functionality of NumPy is imported along with :term:`FiPy`, although
much is augmented for :term:`FiPy`\'s needs.)

>>> if __name__ == "__main__":
...     nx = ny = 1000
... else:
...     nx = ny = 20
>>> mesh = Grid2D(nx=nx, ny=ny, dx=0.25, dy=0.25)
>>> phi = CellVariable(name=r"$\phi$", mesh=mesh)

We start the problem with random fluctuations about :math:`\phi = 1/2`

>>> phi.setValue(GaussianNoiseVariable(mesh=mesh,
...                                    mean=0.5,
...                                    variance=0.01))

:term:`FiPy` doesn't plot or output anything unless you tell it to:

>>> if __name__ == "__main__":
...     viewer = Viewer(vars=(phi,), datamin=0., datamax=1.)

For :term:`FiPy`, we need to perform the partial derivative
:math:`\partial f/\partial \phi`
manually and then put the equation in the canonical
form by decomposing the spatial derivatives
so that each :class:`~fipy.terms.term.Term` is of a single, even order:

.. math::

   \frac{\partial \phi}{\partial t}
    = \nabla\cdot D a^2 \left[ 1 - 6 \phi \left(1 - \phi\right)\right] \nabla \phi- \nabla\cdot D \nabla \epsilon^2 \nabla^2 \phi.

:term:`FiPy` would automatically interpolate
``D * a**2 * (1 - 6 * phi * (1 - phi))``
onto the faces, where the diffusive flux is calculated, but we obtain
somewhat more accurate results by performing a linear interpolation from
``phi`` at cell centers to ``PHI`` at face centers.
Some problems benefit from non-linear interpolations, such as harmonic or
geometric means, and :term:`FiPy` makes it easy to obtain these, too.

>>> PHI = phi.arithmeticFaceValue
>>> D = a = epsilon = 1.
>>> eq = (TransientTerm()
...       == DiffusionTerm(coeff=D * a**2 * (1 - 6 * PHI * (1 - PHI)))
...       - DiffusionTerm(coeff=(D, epsilon**2)))

>>> import fipy.solvers.solver
>>> if fipy.solvers.solver_suite in ['petsc']:
...     solver = DefaultSolver(precon=None)
... elif fipy.solvers.solver_suite in ['trilinos']:
...     solver = LinearLUSolver()
... else:
...     solver = DefaultSolver()

Because the evolution of a spinodal microstructure slows with time, we
use exponentially increasing time steps to keep the simulation
"interesting". The :term:`FiPy` user always has direct control over the
evolution of their problem.

>>> dexp = -5
>>> elapsed = 0.
>>> duration = 1000.

>>> while elapsed < duration:
...     dt = min(100, numerix.exp(dexp))
...     elapsed += dt
...     dexp += 0.01
...     eq.solve(phi, dt=dt, solver=solver)
...     if __name__ == "__main__":
...         viewer.plot()
...     elif (max(phi.globalValue) > 0.7) and (min(phi.globalValue) < 0.3) and elapsed > 10.:
...         break

>>> print((max(phi.globalValue) > 0.7) and (min(phi.globalValue) < 0.3))
True

.. image:: /figures/examples/cahnHilliard/mesh2D.*
   :width: 90%
   :align: center
   :alt: evolution of Cahn-Hilliard phase separation at t = 30, 100 and 1000

"""
__docformat__ = 'restructuredtext'

from fipy import input

if __name__ == '__main__':
    import fipy.tests.doctestPlus
    exec(fipy.tests.doctestPlus._getScript())

    input('finished')



r"""
Test

>>> print((0.4 < max(q.globalValue[0]) < 0.5))
True

"""

__docformat__ = 'restructuredtext'

from fipy import input
from fipy import CellVariable, FaceVariable, Grid1D, TransientTerm, CentralDifferenceConvectionTerm
from fipy.tools import numerix

L = 2.
X0 = -1.
nx = 800
cfl = 0.1
K = 4.
rho = 1.

dx = L / nx
m = Grid1D(nx=nx, dx=dx) + X0
x, = m.cellCenters

q = CellVariable(mesh=m, rank=1, elementshape=(2,))

q[0,:] = numerix.exp(-50 * (x - 0.3)**2) * numerix.cos(20 * (x - 0.3))
q[0, x > 0.3] = 0.

Ax = FaceVariable(mesh=m, rank=3, value=[((0, K), (1 / rho, 0))], elementshape=(1, 2, 2))

eqn = TransientTerm() + CentralDifferenceConvectionTerm(Ax) == 0

if  __name__ == '__main__':
    from fipy import MatplotlibViewer as Viewer
    vi = Viewer((q[0], q[1]))
    vi.plot()

for step in range(500):
    eqn.solve(q, dt=cfl * dx)
    if step % 10 ==  0 and  __name__ == '__main__':
        vi.plot()

if __name__ == '__main__':
    import fipy.tests.doctestPlus
    exec(fipy.tests.doctestPlus._getScript())

    input('finished')




__docformat__ = 'restructuredtext'

from PyTrilinos import AztecOO

from fipy.solvers.trilinos.trilinosAztecOOSolver import TrilinosAztecOOSolver
from fipy.solvers.trilinos.preconditioners.multilevelDDPreconditioner import MultilevelDDPreconditioner

__all__ = ["LinearCGSSolver"]

class LinearCGSSolver(TrilinosAztecOOSolver):

    """Interface to the conjugate gradient squared (:term:`CGS`) solver in
    :ref:`TRILINOS`

    Uses the
    :class:`~fipy.solvers.trilinos.preconditioners.multilevelDDPreconditioner.MultilevelDDPreconditioner`
    by default.
    """

    solver = AztecOO.AZ_cgs

    DEFAULT_PRECONDITIONER = MultilevelDDPreconditioner



##from fipy.tools.profiler.profiler import Profiler
##from fipy.tools.profiler.profiler import calibrate_profiler

"""
Test case for the `Grid3D`.

   >>> eq3D = DiffusionTerm()
   >>> solver = eq3D.getDefaultSolver(tolerance=1e-6)
   >>> eq3D.solve(var, solver=solver)
   >>> eq2D = DiffusionTerm()
   >>> solver = eq2D.getDefaultSolver(tolerance=1e-6)
   >>> eq2D.solve(var2, solver=solver)
   >>> a = numerix.array(var.globalValue)
   >>> b = numerix.array(var2.globalValue)
   >>> c = numerix.ravel(numerix.array((b, b, b)))
   >>> print(numerix.allclose(a, c))
   True

"""

from fipy import input
from fipy import CellVariable, Grid2D, Grid3D, DiffusionTerm, Viewer
from fipy.tools import numerix

nx = 8 # FIXME: downsized temporarily from 10 due to https://github.com/usnistgov/fipy/issues/622
ny = 5
nz = 3

dx = 1.
dy = 1.
dz = 1.

valueBottomTop = 0.
valueLeftRight = 1.

mesh = Grid3D(dx = dx, dy = dy, dz = dz, nx = nx, ny = ny, nz = nz)

var = CellVariable(name = "solution variable",
                   mesh = mesh,
                   value = valueBottomTop)

var.constrain(valueLeftRight, mesh.facesLeft)
var.constrain(valueLeftRight, mesh.facesRight)
var.constrain(valueBottomTop, mesh.facesTop)
var.constrain(valueBottomTop, mesh.facesBottom)

#do the 2D problem for comparison

nx = 8 # FIXME: downsized temporarily from 10 due to https://github.com/usnistgov/fipy/issues/622
ny = 5

dx = 1.
dy = 1.

mesh2 = Grid2D(dx = dx, dy = dy, nx = nx, ny = ny)

var2 = CellVariable(name = "solution variable 2D",
                    mesh = mesh2,
                    value = valueBottomTop)

var2.constrain(valueLeftRight, mesh2.facesLeft)
var2.constrain(valueLeftRight, mesh2.facesRight)
var2.constrain(valueBottomTop, mesh2.facesTop)
var2.constrain(valueBottomTop, mesh2.facesBottom)

eqn = DiffusionTerm()

if __name__ == '__main__':
    eqn.solve(var2)
    viewer = Viewer(var2)
    viewer.plot()
    input("finished")




__docformat__ = 'restructuredtext'

from scipy.sparse.linalg import gmres

from fipy.solvers.scipy.scipyKrylovSolver import ScipyKrylovSolver

__all__ = ["LinearGMRESSolver"]

class LinearGMRESSolver(ScipyKrylovSolver):
    """Interface to the Generalized Minimum RESidual (:term:`GMRES`) solver
    in :ref:`SciPy`.

    No preconditioning by default.
    """

    solveFnc = staticmethod(gmres)

    def _doSolve(self, *args, **kwargs):
        return self.solveFnc(*args, **kwargs, callback_type='legacy')



from fipy.tests.doctestPlus import _LateImportDocTestSuite
import fipy.tests.testProgram

def _suite():
    return _LateImportDocTestSuite(docTestModuleNames = (
            'exponential1D.mesh1D',
            'exponential1D.cylindricalMesh1D',
            'exponential1D.cylindricalMesh1DNonUniform',
            'exponential1D.tri2D',
            'exponential2D.mesh2D',
            'exponential2D.cylindricalMesh2D',
            'exponential2D.cylindricalMesh2DNonUniform',
            'exponential1DBack.mesh1D',
            'powerLaw1D.mesh1D',
            'exponential1DSource.mesh1D',
            'exponential2D.tri2D',
            'exponential1DSource.tri2D',
            'powerLaw1D.tri2D',
            'advection.vanLeerUpwind',
            'peclet',
            'robin',
            'source',
        ), base = __name__)

if __name__ == '__main__':
    fipy.tests.testProgram.main(defaultTest='_suite')



from fipy.solvers.pyamgx import PyAMGXSolver
from fipy.solvers.pyamgx.preconditioners import JacobiPreconditioner

__all__ = ["LinearGMRESSolver"]

class LinearGMRESSolver(PyAMGXSolver):
    """Interface to the Generalized Minimum RESidual (:term:`GMRES`) solver in
    :ref:`PYAMGX`.

    Uses a
    :class:`~fipy.solvers.pyamgx.preconditioners.JacobiPreconditioner` by
    default.
    """

    CONFIG_DICT = {
        "config_version": 2,
        "determinism_flag": 1,
        "exception_handling" : 1,
        "solver": {
            "monitor_residual": 1,
            "solver": "GMRES",
            "convergence": "RELATIVE_INI_CORE",
        }
    }

    DEFAULT_PRECONDITIONER = JacobiPreconditioner



__docformat__ = 'restructuredtext'

from .multilevelPreconditioner import MultilevelPreconditioner

__all__ = ["MultilevelSGSPreconditioner"]

class MultilevelSGSPreconditioner(MultilevelPreconditioner):
    """Multilevel preconditioner using Symmetric Gauss-Seidel smoothing for :class:`~fipy.solvers.trilinos.trilinosSolver.TrilinosSolver`.
    """

    @property
    def _parameterList(self):
        return {
            "output": 0,
            "max levels": self.levels,
            "smoother: type": "symmetric Gauss-Seidel",
        }



import contextlib
import logging
import numpy
import os
from pathlib import Path
import sys
import typer
from typer import Argument, Option
from typing_extensions import Annotated
import unittest
import warnings

from fipy.tests.doctestPlus import _DocTestTimes, report_skips
from fipy.tools.logging.environment import package_info
from fipy.tools import numerix


__all__ = ["app"]

_log = logging.getLogger(__name__)

class DeprecationErroringTestProgram(unittest.TestProgram):
    """`TestProgram` that overrides inability of standard
    `TestProgram` to throw errors on `DeprecationWarning`
    """
    def runTests(self):
        self.warnings = None

        with warnings.catch_warnings():
            warnings.simplefilter(action="error", category=DeprecationWarning)

            # Don't raise noisy errors in
            # tvtk/tvtk_classes.zip/tvtk_classes/abstract_array.py or
            # tvtk/tvtk_classes.zip/tvtk_classes/algorithm.py
            warnings.filterwarnings(action="default", category=DeprecationWarning,
                                    message="invalid escape sequence.*")

            # Don't raise errors in skfmm.pfmm
            # due to deprecation of np.int in NumPy 1.20
            warnings.filterwarnings(action="default", category=DeprecationWarning,
                                    message=r"`np\.int` is a deprecated alias for the builtin `int`.*",
                                    module=r"skfmm\.pfmm")

            # Don't raise errors in scikits.umfpack
            # due to deprecation of pkg_resources.declare_namespace
            # raised in Python 3.7, but not newer
            # (combination of PyTrilinos and Gmsh(?) forces this old configuration)
            warnings.filterwarnings(action="default", category=DeprecationWarning,
                                    message=r"Deprecated call to `pkg_resources\.declare_namespace\('scikits'\)`.*")

            super(DeprecationErroringTestProgram, self).runTests()


class TestCommand(object):
    def __init__(
        self,
        module_or_suite,
        inline,
        pythoncompiled,
        trilinos,
        scipy,
        petsc,
        pyamgx,
        all,
        really_all,
        examples,
        modules,
        viewers,
        cache,
        timetests,
        skfmm,
        lsmlib,
        deprecation_errors,
        verbose
    ):
        self.module_or_suite = module_or_suite
        self.inline = inline
        self.pythoncompiled = pythoncompiled
        self.trilinos = trilinos
        self.scipy = scipy
        self.petsc = petsc
        self.pyamgx = pyamgx
        self.all = all
        self.really_all = really_all
        self.examples = examples
        self.modules = modules
        self.viewers = viewers
        self.cache = cache
        self.timetests = timetests
        self.skfmm = skfmm
        self.lsmlib = lsmlib
        self.deprecation_errors = deprecation_errors
        self.verbose = verbose

        self.initialize()

        self.set_modules()

        self.test_args = list(self._test_args())


    def initialize(self):
        if self.trilinos:
            self.initialize_trilinos()

        if self.pyamgx:
            self.initialize_pyamgx()

        if self.inline:
            self.initialize_weave()

        if self.pythoncompiled is not None:
            pythoncompiled = self.pythoncompiled.absolute().as_posix()
            os.environ['PYTHONCOMPILED'] = pythoncompiled


    def set_modules(self):
        if not (self.examples
                or self.modules
                or self.viewers
                or self.module_or_suite):
            self.all = True
        if self.all or self.really_all:
            self.examples = True
            self.modules = True
        if self.really_all:
            self.viewers = True


    @staticmethod
    @contextlib.contextmanager
    def project_on_sys_path():
        """Add this project to the head of :envvar:`PYTHONPATH`

        Do this in a context that restores the value on exit.

        Cribbed from obsolete :mod:`setuptools.command.test`
        """
        old_path = sys.path[:]
        old_modules = sys.modules.copy()

        try:
            project_path = Path(".").absolute().as_posix()
            sys.path.insert(0, project_path)
            with TestCommand.paths_on_pythonpath([project_path]):
                yield
        finally:
            sys.path[:] = old_path
            sys.modules.clear()
            sys.modules.update(old_modules)


    @staticmethod
    @contextlib.contextmanager
    def paths_on_pythonpath(paths):
        """Add the indicated paths to the head of :envvar:`PYTHONPATH`

        Subprocesses will also see the packages at these paths.

        Do this in a context that restores the value on exit.

        Cribbed from obsolete :mod:`setuptools.command.test`
        """
        nothing = object()
        orig_pythonpath = os.environ.get('PYTHONPATH', nothing)
        current_pythonpath = os.environ.get('PYTHONPATH', '')
        try:
            prefix = os.pathsep.join(paths)
            to_join = filter(None, [prefix, current_pythonpath])
            new_path = os.pathsep.join(to_join)
            if new_path:
                os.environ['PYTHONPATH'] = new_path

            yield
        finally:
            if orig_pythonpath is nothing:
                os.environ.pop('PYTHONPATH', None)
            else:
                os.environ['PYTHONPATH'] = orig_pythonpath


    def _test_args(self):
        # can't seem to delegate a generator until Python 3.3

        if self.verbose:
            yield '--verbose'
        if self.module_or_suite:
            yield self.module_or_suite

        if self.viewers:
            print("*" * 60)
            print("*" + "".center(58) + "*")
            print("*" + "ATTENTION".center(58) + "*")
            print("*" + "".center(58) + "*")
            print("*" + "Some of the following tests require user interaction".center(58) + "*")
            print("*" + "".center(58) + "*")
            print("*" * 60)

            yield "fipy.viewers.testinteractive._suite"
        if self.modules:
            yield "fipy.testFiPy._suite"
        if self.examples:
            yield "examples.test._suite"


    @staticmethod
    def printPackageInfo():
        packages = package_info()

        print()
        package_width = max(len(word) for word in packages.keys())
        for package, version in packages.items():
            print("  ".join((package.ljust(package_width), version)))
        print()


    @staticmethod
    def initialize_trilinos():
        ## The import scipy statement is added to allow
        ## the --Trilinos tests to run without throwing a
        ## segmentation fault. This is caused by weird
        ## behavior in scipy and PyTrilinos depending on
        ## the order in which modules are imported
        try:
            import scipy
        except:
            pass
        import PyTrilinos


    @staticmethod
    def initialize_pyamgx():
        ## Unregister the function pyamgx.finalize
        ## from atexit. This prevents
        ## pyamgx from printing an error message
        ## about memory leaks and a dump of leaked memory.
        ## The memory leaks happen because
        ## the tests do not use the pyamgx solvers
        ## "cleanly", i.e., they do not use the
        ## `with` statement.
        import pyamgx
        import atexit
        if hasattr(atexit, 'unregister'):
            atexit.unregister(pyamgx.finalize)
        else:
            atexit._exithandlers.remove(
                (pyamgx.finalize, (), {}))


    @staticmethod
    def initialize_weave():
        import weave


    @staticmethod
    @contextlib.contextmanager
    def legacy_printoptions():
        printoptions = numerix.get_printoptions()

        try:
            if "legacy" in printoptions:
                numerix.set_printoptions(legacy="1.13")

            yield
        finally:
            if "legacy" in printoptions:
                numerix.set_printoptions(legacy=printoptions["legacy"])


    def save_testtimes(self):
        if self.timetests is None:
            return

        doctesttimes = numpy.rec.fromrecords(_DocTestTimes,
                                formats='f8,S255',
                                names='time,test')
        doctesttimes.sort(order=('time', 'test'))

        # Only write on proc 0.
        # Doesn't use FiPy comms because this command can
        # be run outside of FiPy (`fipy_test`)
        try:
            from mpi4py import MPI
            procID = MPI.COMM_WORLD.rank
            barrier = MPI.COMM_WORLD.barrier
        except:
            procID = 0
            def barrier(*args):
                pass

        if procID == 0:
            numpy.savetxt(self.timetests,
                          doctesttimes[::-1],
                          delimiter='\t',
                          header="time\tmodule",
                          comments='',
                          fmt=("%.18e", "%s"))

        barrier()


    @property
    def test_program_class(self):
        if self.deprecation_errors:
            test_program_class = DeprecationErroringTestProgram
        else:
            test_program_class = unittest.TestProgram

        return test_program_class


    def run_tests(self):
        self.printPackageInfo()

        try:
            # Run tests with current working directory on path
            # so that examples can be found
            with self.project_on_sys_path():
                with self.legacy_printoptions():
                    self.test_program_class(
                        None,
                        None,
                        [unittest.__file__]+self.test_args,
                        testLoader=unittest.TestLoader()
                    )
        except SystemExit as exitErr:
            # unittest.main(..., exit=...) not available until Python 2.7
            report_skips()
            if self.timetests is not None:
                pass
            else:
                raise

        self.save_testtimes()

        if 'exitErr' in locals():
            raise exitErr

def main(
    module_or_suite: Annotated[
        str,
        Argument(help="module or test suite to test")
    ] = None,
    inline: Annotated[
        bool,
        Option(help="run FiPy with inline compilation enabled")
    ] = False,
    pythoncompiled: Annotated[
        Path,
        Option(help="directory in which to put weave's work product")
    ] = None,
    trilinos: Annotated[
        bool,
        Option(help="run FiPy using Trilinos solvers")
    ] = False,
    scipy: Annotated[
        bool,
        Option(help="run FiPy using SciPy solvers")
    ] = False,
    petsc: Annotated[
        bool,
        Option(help="run FiPy using PETSc solvers")
    ] = False,
    pyamgx: Annotated[
        bool,
        Option(help="run FiPy using pyamgx solvers")
    ] = False,
    all: Annotated[
        bool,
        Option(help="run all non-interactive FiPy tests (default)")
    ] = None,
    really_all: Annotated[
        bool,
        Option(help="run *all* FiPy tests (including those requiring user input)")
    ] = False,
    examples: Annotated[
        bool,
        Option(help="test FiPy examples")
    ] = False,
    modules: Annotated[
        bool,
        Option(help="test FiPy code modules")
    ] = False,
    viewers: Annotated[
        bool,
        Option(help="test FiPy viewer modules (requires user input)")
    ] = False,
    cache: Annotated[
        bool,
        Option(help="run FiPy with Variable caching")
    ] = False,
    timetests: Annotated[
        Path,
        Option(help="file in which to put time spent on each test")
    ] = None,
    skfmm: Annotated[
        bool,
        Option(help="run FiPy using the Scikit-fmm level set solver")
    ] = False,
    lsmlib: Annotated[
        bool,
        Option(help="run FiPy using the LSMLIB level set solver")
    ] = False,
    deprecation_errors: Annotated[
        bool,
        Option(help="raise Exceptions for all DeprecationWarnings")
    ] = False,
    verbose: Annotated[
        bool,
        Option(help="blah blah verbose")
    ] = True
):
    test = TestCommand(
        module_or_suite,
        inline,
        pythoncompiled,
        trilinos,
        scipy,
        petsc,
        pyamgx,
        all,
        really_all,
        examples,
        modules,
        viewers,
        cache,
        timetests,
        skfmm,
        lsmlib,
        deprecation_errors,
        verbose
    )

    test.run_tests()


app = typer.Typer()
app.command()(main)

if __name__ == "__main__":
    app()



__docformat__ = 'restructuredtext'

__all__ = []

import os

from fipy.solvers import INDEX_TYPE
from fipy.terms.unaryTerm import _UnaryTerm
from fipy.tools import numerix
from fipy.terms import TermMultiplyError
from fipy.terms import AbstractBaseClassError
from fipy.variables import Variable, FaceVariable, CellVariable

class _AbstractDiffusionTerm(_UnaryTerm):

    def __init__(self, coeff = (1.,), var=None):
        if self.__class__ is _AbstractDiffusionTerm:
            raise AbstractBaseClassError

        self.constraintL = {}
        self.constraintB = {}

        if type(coeff) not in (type(()), type([])):
            coeff = [coeff]

        self.order = len(coeff) * 2


        if len(coeff) > 0:
            self.nthCoeff = coeff[0]

            if not isinstance(self.nthCoeff, Variable):
                self.nthCoeff = Variable(value=self.nthCoeff)

            if isinstance(self.nthCoeff, CellVariable):
                self.nthCoeff = self.nthCoeff.arithmeticFaceValue

        else:
            self.nthCoeff = None

        _UnaryTerm.__init__(self, coeff=coeff, var=var)

        if self.order > 0:
            self.lowerOrderDiffusionTerm = self.__class__(coeff=coeff[1:], var=var)

    def __mul__(self, other):
        if isinstance(other, (int, float)):
            self.coeff[0] = other * self.coeff[0]
            return self.__class__(coeff=self.coeff, var=self.var)
        else:
            raise TermMultiplyError

    __rmul__ = __mul__

    def __neg__(self):
        negatedCoeff = list(self.coeff)

        from fipy.variables.variable import Variable
        if isinstance(negatedCoeff[0], (list, tuple)):
            negatedCoeff[0] = -numerix.array(negatedCoeff[0])
        else:
            negatedCoeff[0] = -negatedCoeff[0]

        return self.__class__(coeff=negatedCoeff, var=self.var)

    def _getBoundaryConditions(self, boundaryConditions):
        higherOrderBCs = []
        lowerOrderBCs = []

        for bc in boundaryConditions:
            bcDeriv = bc._getDerivative(self.order - 2)
            if bcDeriv:
                higherOrderBCs.append(bcDeriv)
            else:
                lowerOrderBCs.append(bc)

        return higherOrderBCs, lowerOrderBCs

    def _getRotationTensor(self, mesh):
        if not hasattr(self, 'rotationTensor'):

            rotationTensor = FaceVariable(mesh=mesh, rank=2)

            rotationTensor[:, 0] = self._getNormals(mesh)

            if mesh.dim == 2:
                rotationTensor[:, 1] = rotationTensor[:, 0].dot(((( 0, 1),
                                                                  (-1, 0))))
            elif mesh.dim == 3:
                epsilon = 1e-20

                div = numerix.sqrt(1 - rotationTensor[2, 0]**2)
                flag = numerix.resize(div > epsilon, (mesh.dim, mesh.numberOfFaces))

                rotationTensor[0, 1] = 1
                rotationTensor[:, 1] = numerix.where(flag,
                                                     rotationTensor[:, 0].dot(((( 0, 1, 0),
                                                                                (-1, 0, 0),
                                                                                ( 0, 0, 0)))) / div,
                                                     rotationTensor[:, 1])


                rotationTensor[1, 2] = 1
                rotationTensor[:, 2] = numerix.where(flag,
                                                     rotationTensor[:, 0] * rotationTensor[2, 0] / div,
                                                     rotationTensor[:, 2])
                rotationTensor[2, 2] = -div

            self.rotationTensor = rotationTensor

        return self.rotationTensor

    def _isotropicOrthogonalCoeff(self, coeff, mesh):
        r"""Geometric coefficient for isotropic diffusion on orthogonal mesh

        Parameters
        ----------
        coeff : array_like
            Diffusion coefficient.
        mesh : ~fipy.meshes.mesh.Mesh
            Geometry and topology.

        Returns
        -------
        ~fipy.variables.faceVariable.FaceVariable
            Contribution to the matrix, consisting of a rank-1 vector at
            each face, composed of the normal contribution.

        Notes
        -----
        .. math::

           \left(\Gamma_0 \frac{A_f}{d_{AP}}\right)^T
        """
        if coeff.shape != () and not isinstance(coeff, FaceVariable):
            # increase dimension of non-scalar coefficient such that
            # it projects to each cell or face it applies to
            coeff = coeff[..., numerix.newaxis]

        return (coeff * FaceVariable(mesh=mesh, value=mesh._faceAreas)
                / mesh._cellDistances)[numerix.newaxis, ...]

    def _anisotropicOrNonorthogonalCoeff(self, coeff, mesh, anisotropicRank):
        r"""Geometric coefficient for anisotropic diffusion or non-orthogonal mesh

        Parameters
        ----------
        coeff : array_like
            Diffusion coefficient.
        mesh : ~fipy.meshes.mesh.Mesh
            Geometry and topology.
        anisotropicRank : int
            ???

        Returns
        -------
        ~fipy.variables.faceVariable.FaceVariable
            Contribution to the matrix, consisting of a rank-1 vector at
            each face, composed of the normal contribution and the
            tangential contribution(s).

        Notes
        -----
        .. math::

           \hat{n}\cdot\Gamma_0\cdot\mathsf{R} A_f

        where :math:`\mathsf{R}` is the rotation tensor.  The normal
        component of the rotation tensor is scaled by the cell distances.
        """
        if anisotropicRank < 2:
            coeff = coeff * numerix.identity(mesh.dim)

        if anisotropicRank > 0:
            shape = numerix.getShape(coeff)
            if mesh.dim != shape[0] or mesh.dim != shape[1]:
                raise IndexError('diffusion coefficient tensor is not an appropriate shape for this mesh')

        faceNormals = FaceVariable(mesh=mesh, rank=1, value=mesh.faceNormals)
        rotationTensor = self._getRotationTensor(mesh)
        rotationTensor[:, 0] = rotationTensor[:, 0] / mesh._cellDistances

        return faceNormals.dot(coeff).dot(rotationTensor) * mesh._faceAreas

    def _calcGeomCoeff(self, var):
        """Geometric coefficient

        Combination of diffusion coefficient and geometric factor.

        Parameters
        ----------
        var : ~fipy.variables.cellVariable.CellVariable
            Solution variable.

        Returns
        -------
        ~fipy.variables.faceVariable.FaceVariable
            Contribution to the matrix, consisting of a rank-1 vector at
            each face, composed of the normal contribution and (for
            nonorthogonal meshes or anisotropic diffusion) the tangential
            contribution(s).
        """

        mesh = var.mesh
        if self.nthCoeff is not None:

            coeff = self.nthCoeff

            if isinstance(coeff, FaceVariable):
                rank = coeff.rank
            else:
                rank = len(numerix.getShape(coeff))

            if var.rank == 0:
                anisotropicRank = rank
            elif var.rank == 1:
                anisotropicRank = rank - 2
            else:
                raise IndexError('the solution variable has the wrong rank')

            if anisotropicRank == 0 and self._treatMeshAsOrthogonal(mesh):

                return self._isotropicOrthogonalCoeff(coeff, mesh)

            else:

                return self._anisotropicOrNonorthogonalCoeff(coeff, mesh, anisotropicRank)

        else:

            return None

    def _getCoefficientMatrix(self, SparseMatrix, var, coeff):
        mesh = var.mesh

        id1, id2 = mesh._adjacentCellIDs
        interiorFaces = numerix.nonzero(mesh.interiorFaces)[0]

        id1 = numerix.take(id1, interiorFaces)
        id2 = numerix.take(id2, interiorFaces)

        id1 = self._reshapeIDs(var, id1)
        id2 = self._reshapeIDs(var, id2)

        facesPerCell = mesh._neighborsPerCell[..., mesh._localNonOverlappingCellIDs]
        coefficientMatrix = self._getMatrix(SparseMatrix=SparseMatrix, mesh=mesh, nonZerosPerRow=facesPerCell + 1)
        interiorCoeff = numerix.take(coeff, interiorFaces, axis=-1).ravel()
        coefficientMatrix.addAt(interiorCoeff, id1.ravel(), id1.swapaxes(0, 1).ravel())
        coefficientMatrix.addAt(-interiorCoeff, id1.ravel(), id2.swapaxes(0, 1).ravel())
        coefficientMatrix.addAt(-interiorCoeff, id2.ravel(), id1.swapaxes(0, 1).ravel())
        coefficientMatrix.addAt(interiorCoeff, id2.ravel(), id2.swapaxes(0, 1).ravel())

        return coefficientMatrix

    def _doBCs(self, SparseMatrix, higherOrderBCs, N, M, coeffs, coefficientMatrix, boundaryB):
        for boundaryCondition in higherOrderBCs:
            LL, bb = boundaryCondition._buildMatrix(SparseMatrix, N, M, coeffs)
            if 'FIPY_DISPLAY_MATRIX' in os.environ:
                self._viewer.title = r"%s %s" % (boundaryCondition.__class__.__name__, self.__class__.__name__)
                self._viewer.plot(matrix=LL, RHSvector=bb)
                from fipy import input
                input()
            coefficientMatrix += LL
            boundaryB += bb

        return coefficientMatrix, boundaryB

    def _constrainValue(self, var):
        r"""Determine value constraint contributions to matrix and RHS

        Parameters
        ----------
        var : ~fipy.variables.cellVariable.CellVariable
            Constrained solution variable

        Returns
        -------
        L : ~fipy.matrices.sparseMatrix.SparseMatrix
            The :math:`N \times N` sparse matrix contribution.
        b : array_like
            The length-N right-hand-side vector contribution.

        Notes
        -----
        For the variable :math:`\phi`, with its value constrained to
        :math:`\phi\rvert_{\partial\Omega_V} = V` on boundary faces
        :math:`\partial\Omega_V`, determines the matrix contribution

        .. math::

             \begin{align}
                \mathsf{L} &= -\nabla\cdot\left(\frac{\Gamma}{d_{fP}}\hat{n}\right)_{f\in\partial\Omega_V} V_P
                \\
                &\approx -\sum_{f\in\partial\Omega_V}(\frac{\Gamma}{d_{fP}}\hat{n}\cdot\hat{n})_f A_f
             \end{align}

        and the right-hand-side vector contribution

        .. math::

             \begin{align}
                \mathbf{b} &= -\nabla\cdot\left(\frac{\Gamma V}{d_{fP}}\hat{n}\right)_{f\in\partial\Omega_V} V_P
                \\
                &\approx -\sum_{f\in\partial\Omega_V}(\frac{\Gamma V}{d_{fP}}\hat{n}\cdot\hat{n})_f A_f
             \end{align}
        """
        mesh = var.mesh
        normals = FaceVariable(mesh=mesh, rank=1, value=mesh._orientedFaceNormals)

        if len(var.shape) == 1 and len(self.nthCoeff.shape) > 1:
            normalsNthCoeff =  normals.dot(self.nthCoeff)
        else:

            if self.nthCoeff.shape != () and not isinstance(self.nthCoeff, FaceVariable):
                coeff = self.nthCoeff[..., numerix.newaxis]
            else:
                coeff = self.nthCoeff

            s = (slice(0, None, None),) + (numerix.newaxis,) * (len(coeff.shape) - 1) + (slice(0, None, None),)
            normalsNthCoeff = coeff[numerix.newaxis] * normals[s]

        constrainedNormalsDotCoeffOverdAP = var.arithmeticFaceValue.constraintMask * \
                                            normalsNthCoeff / mesh._cellDistances

        L = -constrainedNormalsDotCoeffOverdAP.divergence * mesh.cellVolumes
        b = -(constrainedNormalsDotCoeffOverdAP
              * var.arithmeticFaceValue).divergence * mesh.cellVolumes

        return L, b

    def _constrainGradient(self, var):
        r"""Determine gradient constraint contributions to matrix and RHS

        Parameters
        ----------
        var : ~fipy.variables.cellVariable.CellVariable
            Constrained solution variable of N cells.

        Returns
        -------
        L : ~fipy.matrices.sparseMatrix.SparseMatrix
            The :math:`N \times N` sparse matrix contribution.
        b : array_like
            The length-N right-hand-side vector contribution

        Notes
        -----
        For the variable :math:`\phi`, with its gradient constrained to
        :math:`\nabla\phi\rvert_{\partial\Omega_G} = \vec{G}` on boundary
        faces :math:`\partial\Omega_G`, determines the matrix contribution

        .. math::

             \begin{align}
                \mathsf{L} &= \mathsf{0}
             \end{align}

        and the right-hand-side vector contribution

        .. math::

             \begin{align}
                 \mathbf{b} &= -\nabla\cdot\left(\Gamma\vec{G}\right)_{f\in\partial\Omega_G} V_P
                 \\
                 &\approx -\sum_{f\in\partial\Omega_G}(\Gamma\vec{G}\cdot\hat{n})_f A_f
             \end{align}
        """
        if len(var.shape) == 1 and len(self.nthCoeff.shape) > 1:
            # var is scalar field and self.nthCoeff is vector (or tensor)
            nthCoeffFaceGrad = var.faceGrad.dot(self.nthCoeff)
        else:
            # var is vector or tensor field or self.nthCoeff is scalar
            if not (self.nthCoeff.shape == () or isinstance(self.nthCoeff, FaceVariable)):
                # self.nthCoeff is not a scalar or a FaceVariable
                coeff = self.nthCoeff[..., numerix.newaxis]
            else:
                # self.nthCoeff is a scalar or a FaceVariable
                coeff = self.nthCoeff

            nthCoeffFaceGrad = coeff[numerix.newaxis] * var.faceGrad[:, numerix.newaxis]

        b = -(var.faceGrad.constraintMask
              * nthCoeffFaceGrad).divergence * var.mesh.cellVolumes

        return 0, b


    def _calcConstraints(self, var):
        """Determine contributions to matrix and RHS due to constraints on `var`

        Parameters
        ----------
        var : ~fipy.variables.cellVariable.CellVariable
            The constrained variable

        Returns
        -------
        None
        """
        if (var not in self.constraintL) or (var not in self.constraintB):
            LL, bb = self._constrainValue(var)

            self.constraintL[var] = LL
            self.constraintB[var] = bb

            LL, bb = self._constrainGradient(var)

            self.constraintL[var] += LL
            self.constraintB[var] += bb

    def _buildMatrix(self, var, SparseMatrix, boundaryConditions=(), dt=None, transientGeomCoeff=None, diffusionGeomCoeff=None):
        """

        Test to ensure that a changing coefficient influences the boundary conditions.

        >>> from fipy import *
        >>> m = Grid2D(nx=2, ny=2)
        >>> v = CellVariable(mesh=m)
        >>> c0 = Variable(1.)
        >>> v.constrain(c0, where=m.facesLeft)

        Diffusion will only be in the y-direction

        >>> coeff = Variable([[0., 0.], [0., 1.]])
        >>> eq = DiffusionTerm(coeff)
        >>> eq.solve(v, solver=DummySolver())
        >>> print(v)
        [ 0.  0.  0.  0.]

        Change the coefficient.

        >>> coeff[0, 0] = 1.
        >>> eq.solve(v)
        >>> print(v)
        [ 1.  1.  1.  1.]

        Change the constraints.

        >>> c0.setValue(2.)
        >>> v.constrain(3., where=m.facesRight)
        >>> print(v.faceValue.constraintMask)
        [False False False False False False  True False  True  True False  True]
        >>> eq.solve(v)
        >>> print(v)
        [ 2.25  2.75  2.25  2.75]

        """

        if self.order > 2:
            buildFn = self._higherOrderBuildMatrix
        elif self.order == 2:
            buildFn = self._secondOrderBuildMatrix
        elif self.order == 0:
            buildFn = self._zerothOrderBuildMatrix
        else:
            raise ValueError("Order of diffusion coefficient must be non-zero and even, not {}".format(self.order))

        return buildFn(var=var,
                       SparseMatrix=SparseMatrix,
                       boundaryConditions=boundaryConditions,
                       dt=dt,
                       transientGeomCoeff=transientGeomCoeff,
                       diffusionGeomCoeff=diffusionGeomCoeff)

    def _calcAnisotropySource(self, coeff, var):

        if not hasattr(self, 'anisotropySource'):
            if len(coeff) > 1:
                mesh = var.mesh
                unconstrainedVar = var + 0
                gradients = unconstrainedVar.grad.harmonicFaceValue.dot(self._getRotationTensor(mesh))
                from fipy.variables.addOverFacesVariable import _AddOverFacesVariable
                self.anisotropySource = _AddOverFacesVariable(gradients[1:].dot(coeff[1:])) * mesh.cellVolumes

    def _calcCoeffDict(self, var):
        """Matrix contributions to cells on either side of face

        Returns
        -------
        dict
            .. table:: Stencil contributions
               :widths: auto

               ==============  ============
                     key           value
               ==============  ============
               `cell1diag`     `-geomCoeff`
               `cell1offdiag`  `geomCoeff`
               `cell2offdiag`  `geomCoeff`
               `cell2diag`     `-geomCoeff`
               ==============  ============

        Notes
        -----
        For 2nd-order diffusion, if the diffusion coefficient is
        anisotropic or the mesh is nonorthogonal, also sets
        :prop:`~fipy.terms.abstractDiffusionTerm._AbstractDiffusionTerm.anisotropySource`.
        """
        if not hasattr(self, 'coeffDict'):

            coeff = self._getGeomCoeff(var)

            coeffDict = {
                'cell 1 diag':     -coeff[0],
                'cell 1 offdiag':  coeff[0]
                }

            coeffDict['cell 1 diag'].dontCacheMe()
            coeffDict['cell 1 offdiag'].dontCacheMe()

            coeffDict['cell 2 offdiag'] = coeffDict['cell 1 offdiag']
            coeffDict['cell 2 diag'] = coeffDict['cell 1 diag']

            self.coeffDict = coeffDict

            if self.order == 2:
                self._calcAnisotropySource(coeff, var)

            del coeff

    def _higherOrderBuildMatrix(self, var, SparseMatrix, boundaryConditions=(), dt=None, transientGeomCoeff=None, diffusionGeomCoeff=None):
        r"""Recursively build matrix and RHS vector

        Parameters
        ----------
        var : ~fipy.variables.cellVariable.CellVariable
            Solution variable of N cells.
        SparseMatrix : class
            ~fipy.matrices.sparseMatrix.SparseMatrix class to build into.
        boundaryConditions : tuple of ~fipy.boundaryConditions.boundaryCondition.BoundaryCondition
            Old-style (pre-contraint) boundary conditions to apply.
        dt : float
            Time step.
        transientGeomCoeff : ~fipy.variables.cellVariable.CellVariable
            Unused.
        diffusionGeomCoeff : ~fipy.variables.faceVariable.FaceVariable
            Unused.

        Returns
        -------
        var : ~fipy.variables.cellVariable.CellVariable
            Solution variable of N cells.
            Why pass in `var` and then pass it back out?
        L : ~fipy.matrices.sparseMatrix.SparseMatrix
            The :math:`N \times N` sparse matrix from this and all lower-order
            contributions.
        b : array_like
            The length-N right-hand-side vector from this and all
            lower-order contributions.

        Notes
        -----
        Given an :math:`O^{th}`-order diffusion term

        .. math::

             \nabla\cdot\{\Gamma_2 \nabla [\nabla\cdot(\Gamma_4 \nabla \cdots \{ \Gamma_O \nabla \phi \})]\}

        recursively determines the matrix

        .. math::

             \mathsf{L} = \mathsf{L}_O \mathsf{L}_{O-2} \frac{1}{V_P} \mathsf{I}

        and the right-hand-side vector

        .. math::

             \mathbf{b} = \mathbf{b}_O + \mathsf{L}_O \mathbf{b}_{O-2} \frac{1}{V_P}
        """
        mesh = var.mesh

        N = mesh.numberOfCells
        M = mesh._maxFacesPerCell

        higherOrderBCs, lowerOrderBCs = self._getBoundaryConditions(boundaryConditions)

        (var,
         lowerOrderL,
         lowerOrderb) = self.lowerOrderDiffusionTerm._buildMatrix(var=var,
                                                                  SparseMatrix=SparseMatrix,
                                                                  boundaryConditions=lowerOrderBCs,
                                                                  dt=dt,
                                                                  transientGeomCoeff=transientGeomCoeff,
                                                                  diffusionGeomCoeff=diffusionGeomCoeff)
        del lowerOrderBCs

        lowerOrderb = lowerOrderb / mesh.cellVolumes
        volMatrix = SparseMatrix(mesh=var.mesh, nonZerosPerRow=1)

        volMatrix.addAtDiagonal(1. / mesh.cellVolumes)
        lowerOrderL = volMatrix * lowerOrderL
        del volMatrix

        self._calcCoeffDict(var)

        L = self._getCoefficientMatrix(SparseMatrix, var, self.coeffDict['cell 1 diag'])
        L, b = self._doBCs(SparseMatrix, higherOrderBCs, N, M, self.coeffDict,
                           L, numerix.zeros(len(var.ravel()), 'd'))

        del higherOrderBCs

        b = numerix.asarray(L * lowerOrderb) + b
        del lowerOrderb

        L = L * lowerOrderL
        del lowerOrderL

        return (var, L, b)

    def _secondOrderBuildMatrix(self, var, SparseMatrix, boundaryConditions=(), dt=None, transientGeomCoeff=None, diffusionGeomCoeff=None):
        r"""Build the 2nd-order matrix and RHS vector

        .. math::

           \nabla\cdot(\Gamma_0 \nabla \phi)

        Parameters
        ----------
        var : ~fipy.variables.cellVariable.CellVariable
            Solution variable of N cells.
        SparseMatrix : class
            ~fipy.matrices.sparseMatrix.SparseMatrix class to build into.
        boundaryConditions : tuple of ~fipy.boundaryConditions.boundaryCondition.BoundaryCondition
            Old-style (pre-contraint) boundary conditions to apply.
        dt : float
            Time step.
        transientGeomCoeff : ~fipy.variables.cellVariable.CellVariable
            Unused.
        diffusionGeomCoeff : ~fipy.variables.faceVariable.FaceVariable
            Unused.

        Returns
        -------
        var : ~fipy.variables.cellVariable.CellVariable
            Solution variable of N cells.
            Why pass in `var` and then pass it back out?
        L : ~fipy.matrices.sparseMatrix.SparseMatrix
            The :math:`N \times N` sparse matrix from this second-order contribution.
        b : array_like
            Right-hand-side vector from this second-order contribution.

        Notes
        -----
        Given a 2nd-order diffusion term

        .. math::

             \nabla\cdot(\Gamma_0 \nabla \phi)

        determines the matrix

        .. math::

             \mathsf{L} = \mathsf{L}_2

        and the right-hand-side vector

        .. math::

             \mathbf{b} = \mathbf{b}_2
        """
        mesh = var.mesh

        N = mesh.numberOfCells
        M = mesh._maxFacesPerCell

        higherOrderBCs, lowerOrderBCs = self._getBoundaryConditions(boundaryConditions)
        del lowerOrderBCs

        self._calcCoeffDict(var)

        L = self._getCoefficientMatrix(SparseMatrix, var, self.coeffDict['cell 1 diag'])
        L, b = self._doBCs(SparseMatrix, higherOrderBCs, N, M, self.coeffDict,
                           L, numerix.zeros(len(var.ravel()), 'd'))

        del higherOrderBCs

        if hasattr(self, 'anisotropySource'):
            b -= self.anisotropySource

        self._calcConstraints(var)

        ids = self._reshapeIDs(var, numerix.arange(mesh.numberOfCells,
                                                   dtype=INDEX_TYPE))
        L.addAt(self.constraintL[var].ravel(),
                ids.ravel(),
                ids.swapaxes(0, 1).ravel())
        b += numerix.reshape(self.constraintB[var].ravel(),
                             ids.shape).sum(-2).ravel()

        return (var, L, b)

    def _zerothOrderBuildMatrix(self, var, SparseMatrix, boundaryConditions=(), dt=None, transientGeomCoeff=None, diffusionGeomCoeff=None):
        r"""Recursively build the 0th-order matrix and RHS vector

        The purpose of diffusion order 0 is to enable recursive
        construction of higher-order diffusion terms.

        Parameters
        ----------
        var : ~fipy.variables.cellVariable.CellVariable
            Solution variable of N cells.
        SparseMatrix : class
            ~fipy.matrices.sparseMatrix.SparseMatrix class to build into.
        boundaryConditions : tuple of ~fipy.boundaryConditions.boundaryCondition.BoundaryCondition
            Unused.
        dt : float
            Unused.
        transientGeomCoeff : ~fipy.variables.cellVariable.CellVariable
            Unused.
        diffusionGeomCoeff : ~fipy.variables.faceVariable.FaceVariable
            Unused.

        Returns
        -------
        var : ~fipy.variables.cellVariable.CellVariable
            Solution variable of N cells.
            Why pass in `var` and then pass it back out?
        L : ~fipy.matrices.sparseMatrix.SparseMatrix
            The :math:`N \times N` sparse matrix from this zeroth-order contribution.
        b : array_like
            Right-hand-side vector from this zeroth-order contribution.

        Notes
        -----
        Given a 0th-order diffusion term

        .. math::

             \phi

        determines the matrix

        .. math::

             \mathsf{L} = V_P \mathsf{I}

        and the right-hand-side vector

        .. math::

             \mathbf{b} = \mathbf{0}
        """
        mesh = var.mesh

        N = mesh.numberOfCells
        M = mesh._maxFacesPerCell

        L = SparseMatrix(mesh=mesh, nonZerosPerRow=1)
        L.addAtDiagonal(mesh.cellVolumes)
        b = numerix.zeros(len(var.ravel()), 'd')

        return (var, L, b)

    def _getDiffusionGeomCoeff(self, var):
        if var is self.var or self.var is None:
            return self._getGeomCoeff(var)
        else:
            return None

    @property
    def _diffusionVars(self):
        return self._vars

    def _treatMeshAsOrthogonal(self, mesh):
        raise NotImplementedError

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()





__docformat__ = 'restructuredtext'

from fipy.terms.cellTerm import CellTerm
from fipy.terms import AbstractBaseClassError
from fipy.variables.cellVariable import CellVariable
from fipy.tools import numerix

__all__ = ["SourceTerm"]

class SourceTerm(CellTerm):
    """
    .. attention:: This class is abstract. Always create one of its subclasses.
    """
    def __init__(self, coeff=0., var=None):
        r"""
        Parameters
        ----------
        coeff : float or ~fipy.variables.cellVariable.CellVariable
            Coefficient of source (default: 0)
        var : ~fipy.variables.cellVariable.CellVariable
            Variable :math:`\phi` that
            :class:`~fipy.terms.sourceTerm.SourceTerm` is implicit in.
        """
        if self.__class__ is SourceTerm:
            raise AbstractBaseClassError
        super(SourceTerm, self).__init__(coeff=coeff, var=var)

    def _calcGeomCoeff(self, var):
        self._checkCoeff(var)

        if self.coeff.shape != () and self.coeff.shape[-1] != len(var.mesh.cellVolumes):
            return self.coeff[..., numerix.newaxis] * CellVariable(mesh=var.mesh, value=var.mesh.cellVolumes)
        else:
            return self.coeff * CellVariable(mesh=var.mesh, value=var.mesh.cellVolumes)

    def _checkDt(self, dt):
        return 1.




import logging

_log = logging.getLogger(__name__)

from .preconditioners import *

from .linearCGSSolver import *
from .linearGMRESSolver import *
from .linearBicgstabSolver import *
from .linearLUSolver import *
from .linearCGSolver import *

from . import scipyConvergence

DefaultSolver = LinearLUSolver
DefaultAsymmetricSolver = LinearLUSolver
DummySolver = LinearGMRESSolver
GeneralSolver = LinearLUSolver



__docformat__ = 'restructuredtext'

from fipy.variables.cellVariable import CellVariable
from fipy.variables.distanceVariable import DistanceVariable

class GapFillDistanceVariable(DistanceVariable):

    def extendVariable(self, extensionVariable, order=2):
        if not hasattr(self, 'fineDistanceVariable'):
            self.fineDistanceVariable = DistanceVariable(mesh=self.mesh.fineMesh)
        if not hasattr(self, 'fineExtensionVariable'):
            self.fineExtensionVariable = CellVariable(mesh=self.mesh.fineMesh)
        self.fineDistanceVariable[:] = self(self.mesh.fineMesh.cellCenters)
        self.fineExtensionVariable[:] = extensionVariable(self.mesh.fineMesh.cellCenters)
        self.fineDistanceVariable.extendVariable(self.fineExtensionVariable, order=order)
        extensionVariable[:] = self.fineExtensionVariable(self.mesh.cellCenters)

    def calcDistanceFunction(self, order=2):
        if not hasattr(self, 'fineDistanceVariable'):
            self.fineDistanceVariable = DistanceVariable(mesh=self.mesh.fineMesh)
        self.fineDistanceVariable[:] = self(self.mesh.fineMesh.cellCenters)
        self.fineDistanceVariable.calcDistanceFunction(order=order)
        self[:] = self.fineDistanceVariable(self.mesh.cellCenters)

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()



__docformat__ = 'restructuredtext'

from fipy.solvers.petsc.petscKrylovSolver import PETScKrylovSolver

__all__ = ["LinearCGSSolver"]

class LinearCGSSolver(PETScKrylovSolver):

    """Interface to the conjugate gradient squared (:term:`CGS`) solver in
    :ref:`PETSc`.
    """
      
    solver = 'cgs'



import os
import sys
import re
from subprocess import Popen, PIPE

from fipy.tools.parser import parse

from examples.benchmarking.utils import monitor

def main():
    steps = parse('--numberOfSteps', action='store',
                  type='int', default=20)

    blocks = parse('--numberOfBlocks', action='store',
                   type='int', default=20)

    benchmarker = os.path.join(os.path.dirname(__file__),
                               "benchmarker.py")

    args = sys.argv[1:]

    pyth = sys.executable or "python"

    p = Popen([pyth, benchmarker] + args
              + ["--numberOfSteps=0"],
              stdout=PIPE,
              stderr=PIPE)

    cpu0, rsz0, vsz0 = monitor(p)

    print("step\tcpu / (s / step / cell)\trsz / (B / cell)\tvsz / (B / cell)")

    for block in range(blocks):
        p = Popen([pyth, benchmarker,
                   "--startingStep=%d" % (block * steps),
                   "--cpuBaseLine=%f" % cpu0] + args,
                  stdout=PIPE,
                  stderr=PIPE)

        cpu, rsz, vsz = monitor(p)

        print("%d\t%g\t%g\t%g" % (block * steps, cpu, rsz, vsz))

if __name__ == "__main__":
    main()



"""Boundary conditions
"""
__docformat__ = 'restructuredtext'

from fipy.boundaryConditions.constraint import *
from fipy.boundaryConditions.fixedFlux import *
from fipy.boundaryConditions.fixedValue import *
from fipy.boundaryConditions.nthOrderBoundaryCondition import *



r"""Solve a fourth-order diffusion problem.

This example uses the :class:`~fipy.terms.diffusionTerm.DiffusionTerm` class to solve the equation

.. math::

   \frac{\partial^4 \phi}{\partial x^4} = 0

on a 1D mesh of length

>>> L = 1000.

We create an appropriate mesh

.. index::
   single: Grid1D

>>> from fipy import CellVariable, Grid1D, NthOrderBoundaryCondition, DiffusionTerm, Viewer, LinearLUSolver

>>> nx = 500
>>> dx = L / nx
>>> mesh = Grid1D(dx=dx, nx=nx)

and initialize the solution variable to 0

.. index::
   single: CellVariable

>>> var = CellVariable(mesh=mesh, name='solution variable')

For this problem, we impose the boundary conditions:

.. math::

   \phi &= \alpha_1 \quad \text{at $x = 0$} \\
   \frac{\partial \phi}{\partial x} &= \alpha_2 \quad \text{at $x = L$} \\
   \frac{\partial^2 \phi}{\partial x^2} &= \alpha_3 \quad \text{at $x = 0$} \\
   \frac{\partial^3 \phi}{\partial x^3} &= \alpha_4 \quad \text{at $x = L$.}

or

>>> alpha1 = 2.
>>> alpha2 = 1.
>>> alpha3 = 4.
>>> alpha4 = -3.

.. index::
   single: NthOrderBoundaryCondition

>>> BCs = (NthOrderBoundaryCondition(faces=mesh.facesLeft, value=alpha3, order=2),
...        NthOrderBoundaryCondition(faces=mesh.facesRight, value=alpha4, order=3))
>>> var.faceGrad.constrain([alpha2], mesh.facesRight)
>>> var.constrain(alpha1, mesh.facesLeft)

We initialize the steady-state equation

>>> eq = DiffusionTerm(coeff=(1, 1)) == 0

>>> solver = LinearLUSolver()

We perform one implicit timestep to achieve steady state

>>> eq.solve(var=var,
...          boundaryConditions=BCs,
...          solver=solver)

The analytical solution is:

.. math::

   \phi = \frac{ \alpha_4 }{6} x^3 + \frac{ \alpha_3 }{2} x^2
   + \left( \alpha_2 - \frac{ \alpha_4 }{2} L^2  - \alpha_3 L \right) x + \alpha_1

or

>>> analytical = CellVariable(mesh=mesh, name='analytical value')
>>> x = mesh.cellCenters[0]
>>> analytical.setValue(alpha4 / 6. * x**3 + alpha3 / 2. * x**2 + \
...                     (alpha2 - alpha4 / 2. * L**2 - alpha3 * L) * x + alpha1)

>>> print(var.allclose(analytical, rtol=1e-4))
1

If the problem is run interactively, we can view the result:

.. index::
   pair: module; fipy.viewers

>>> if __name__ == '__main__':
...     viewer = Viewer(vars=(var, analytical))
...     viewer.plot()

.. image:: /figures/examples/diffusion/input4thOrder1D.*
   :width: 90%
   :align: center
   :alt: solution to biharmonic equation

"""
__docformat__ = 'restructuredtext'

from fipy import input

if __name__ == '__main__':
    import fipy.tests.doctestPlus
    exec(fipy.tests.doctestPlus._getScript())
    input('finished')




__docformat__ = 'restructuredtext'

from fipy.terms.implicitSourceTerm import ImplicitSourceTerm
from fipy.variables.levelSetDiffusionVariable import _LevelSetDiffusionVariable
from fipy.terms.transientTerm import TransientTerm
from fipy.terms.diffusionTerm import DiffusionTermNoCorrection

def buildSurfactantBulkDiffusionEquation(bulkVar = None,
                                         distanceVar = None,
                                         surfactantVar = None,
                                         otherSurfactantVar = None,
                                         diffusionCoeff = None,
                                         transientCoeff = 1.,
                                         rateConstant = None):
    r"""

    The `buildSurfactantBulkDiffusionEquation` function returns a bulk diffusion of a
    species with a source term for the jump from the bulk to an interface.
    The governing equation is given by,

    .. math::

       \frac{\partial c}{\partial t} = \nabla \cdot D \nabla  c

    where,

    .. math::

       D = \begin{cases}
           D_c & \text{when $\phi > 0$} \\
           0  & \text{when $\phi \le 0$}
       \end{cases}

    The jump condition at the interface is defined by Langmuir
    adsorption. Langmuir adsorption essentially states that the ability for
    a species to jump from an electrolyte to an interface is proportional to
    the concentration in the electrolyte, available site density and a
    jump coefficient. The boundary condition at the interface is given by

    .. math::

       D \hat{n} \cdot \nabla c = -k c (1 - \theta) \qquad \text{at $\phi = 0$}.

    Parameters
    ----------
    bulkVar : ~fipy.variables.cellVariable.CellVariable
        The bulk surfactant concentration variable.
    distanceVar : ~fipy.variables.distanceVariable.DistanceVariable
    surfactantVar : ~fipy.variables.surfactantVariable.SurfactantVariable
    otherSurfactantVar : ~fipy.variables.surfactantVariable.SurfactantVariable
        Any other surfactants that may remove this one.
    diffusionCoeff : float or ~fipy.variables.faceVariable.FaceVariable
    transientCoeff : float
        In general 1 is used.
    rateConstant : float
        The adsorption coefficient.

    """

    spCoeff = rateConstant * distanceVar.cellInterfaceAreas / bulkVar.mesh.cellVolumes
    spSourceTerm = ImplicitSourceTerm(spCoeff)

    bulkSpCoeff = spCoeff * bulkVar
    coeff = bulkSpCoeff * surfactantVar.interfaceVar

    diffusionCoeff = _LevelSetDiffusionVariable(distanceVar,
                                                diffusionCoeff)

    eq =  TransientTerm(transientCoeff) - DiffusionTermNoCorrection(diffusionCoeff)

    if otherSurfactantVar is not None:
        otherCoeff = bulkSpCoeff * otherSurfactantVar.interfaceVar
    else:
        otherCoeff = 0

    return eq - coeff + spSourceTerm - otherCoeff



__docformat__ = 'restructuredtext'

from .petscPreconditioner import PETScPreconditioner

__all__ = ["SSORPreconditioner"]

class SSORPreconditioner(PETScPreconditioner):
    """
    SSOR preconditioner for :class:`~fipy.solvers.petsc.petscSolver.PETScSolver`.

    """

    pctype = "sor"



r"""

This example solves the steady-state convection-diffusion equation as described in
:mod:`examples.convection.exponential1D.mesh1D` but uses a
:class:`~fipy.meshes.tri2D.Tri2D` mesh.

Here the axes are reversed (``nx = 1``, ``ny = 1000``) and

.. math::

   \vec{u} = (0, 10)

>>> from fipy import CellVariable, Tri2D, DiffusionTerm, ExponentialConvectionTerm, DefaultAsymmetricSolver, Viewer
>>> from fipy.tools import numerix

>>> L = 10.
>>> nx = 1
>>> ny = 1000
>>> mesh = Tri2D(dx = L / ny, dy = L / ny, nx = nx, ny = ny)

>>> valueBottom = 0.
>>> valueTop = 1.

>>> var = CellVariable(name = "concentration",
...                    mesh = mesh,
...                    value = valueBottom)

>>> var.constrain(valueBottom, mesh.facesBottom)
>>> var.constrain(valueTop, mesh.facesTop)

>>> diffCoeff = 1.
>>> convCoeff = numerix.array(((0.,), (10.,)))

>>> eq = (DiffusionTerm(coeff=diffCoeff)
...       + ExponentialConvectionTerm(coeff=convCoeff))

It should be possible to drive this steady-state solution to high accuracy.

>>> solver = DefaultAsymmetricSolver(tolerance=1e-10, iterations=10000)
>>> eq.solve(var=var, solver=solver)

The analytical solution test for this problem is given by:

>>> axis = 1
>>> y = mesh.cellCenters[axis]
>>> CC = 1. - numerix.exp(-convCoeff[axis] * y / diffCoeff)
>>> DD = 1. - numerix.exp(-convCoeff[axis] * L / diffCoeff)
>>> analyticalArray = CC / DD

>>> print(var.allclose(analyticalArray, rtol=1e-6, atol=1e-6))
True

>>> if __name__ == '__main__':
...     viewer = Viewer(vars = var)
...     viewer.plot()
"""
__docformat__ = 'restructuredtext'

from fipy import input

if __name__ == '__main__':
    import fipy.tests.doctestPlus
    exec(fipy.tests.doctestPlus._getScript())

    input('finished')




__docformat__ = 'restructuredtext'

from scipy.sparse.linalg import cg

from fipy.solvers.scipy.scipyKrylovSolver import ScipyKrylovSolver

__all__ = ["LinearCGSolver", "LinearPCGSolver"]

class LinearCGSolver(ScipyKrylovSolver):
    """Interface to the conjugate gradient (:term:`CG`) solver
    in :ref:`Scipy`.

    No preconditioning by default.
    """

    solveFnc = staticmethod(cg)

    def _canSolveAsymmetric(self):
        return False

LinearPCGSolver = LinearCGSolver



import logging

_log = logging.getLogger(__name__)

from .linearLUSolver import *
from .linearCGSolver import *
from .linearGMRESSolver import *
from .linearBicgSolver import *
from .linearCGSSolver import *
from .dummySolver import *
from . import petscConvergence

from .preconditioners import *

DefaultSolver = LinearGMRESSolver
DefaultAsymmetricSolver = LinearGMRESSolver
DummySolver = DummySolver
GeneralSolver = DefaultSolver
